﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Windows.Shapes;
using UmlDiagramDesigner.Core.Enums;
using UmlDiagramDesigner.Properties;
using UmlDiagramDesigner.ViewModels.DiagramItems;
using Point = System.Windows.Point;
using Rectangle = System.Drawing.Rectangle;

namespace UmlDiagramDesigner.Helpers
{
    public static class ExtensionMethods
    {
        public static bool IsPartOf(this Point point, Rectangle rectangle)
        {
            return rectangle.IntersectsWith(new Rectangle((int) point.X, (int) point.Y, 1, 1));
        }

        /// <summary>
        /// Determines whether the point is a part of the rectangle.
        /// </summary>
        /// <param name="point">The point.</param>
        /// <param name="rectangles">The rectangles.</param>
        /// <returns></returns>
        public static bool IsPartOfAny(this Point point, IEnumerable<Rectangle> rectangles)
        {
			if(rectangles == null) throw new ArgumentNullException("rectangles");
            return
                rectangles.Any(rectangle => rectangle.IntersectsWith(new Rectangle((int) point.X, (int) point.Y, 1, 1)));
        }


        /// <summary>
        /// Gets the items which intersect with the point (based on the items Rectangle property).
        /// </summary>
        /// <param name="items">The items.</param>
        /// <param name="point">The point.</param>
        /// <returns></returns>
        public static IEnumerable<DiagramItemViewModelBase> GetIntersectingItems(this IEnumerable<DiagramItemViewModelBase> items,
            Point point)
        {
			if(items == null) throw new ArgumentNullException("items");
            return items.Where(item => point.IsPartOf(item.Rectangle));
        }

        /// <summary>
        /// Gets the top intersecting item based on it's ZIndex value.
        /// </summary>
        /// <param name="items">The items.</param>
        /// <param name="point">The point.</param>
        /// <returns></returns>
        public static DiagramItemViewModelBase GetTopIntersectingItem(this IEnumerable<DiagramItemViewModelBase> items,
    Point point)
        {
            return items.Where(item => point.IsPartOf(item.Rectangle)).OrderByDescending(item => item.ZIndex).FirstOrDefault();
        }


        /// <summary>
        /// Calculates the distance from source point to other one.
        /// </summary>
        /// <param name="source">The source.</param>
        /// <param name="other">The other.</param>
        /// <returns></returns>
        public static double CalculateDistance(this Point source, Point other)
        {
            var dx = source.X - other.X;
            var dy = source.Y - other.Y;
            return (dx*dx )+ (dy*dy);
        }


        /// <summary>
        /// Returns a collection of rectangle's points around its circumference with specified step between them starting from middle out.
        /// If you want to return all points, set step to 1.
        /// </summary>
        /// <param name="rectangle">The rectangle.</param>
        /// <param name="stepValue"></param>
        /// <returns></returns>
        public static IEnumerable<Point> GetPoints(this Rectangle rectangle, int? stepValue = null)
        {
	        int step = Settings.Default.PathFindingSensitivity;

            if (stepValue.HasValue && stepValue.Value > 0)
                step = stepValue.Value;

            var points = new List<Point>();

            if (rectangle.Width < step || rectangle.Width <= 0) 
            {
                points.Add(new Point(rectangle.Location.X, rectangle.Location.Y));
                points.Add(new Point(rectangle.Location.X + rectangle.Width, rectangle.Y));
            }
            if (rectangle.Height < step || rectangle.Height <= 0)
            {
                points.Add(new Point(rectangle.Location.X, rectangle.Location.Y + rectangle.Width));
                points.Add(new Point(rectangle.X + rectangle.Width, rectangle.Y + rectangle.Height));
            }
            if (points.Count == 4) return points;

            int modulo = rectangle.Width % step;

            for (int i = rectangle.Location.X; i < rectangle.Location.X + rectangle.Width; i += step)
            {
                points.Add(new Point(i, rectangle.Location.Y));
                points.Add(new Point(i, rectangle.Location.Y + rectangle.Height));
            }
            if (modulo != 0) points.Add(new Point(rectangle.Location.X + rectangle.Width, rectangle.Location.Y));

            for (int i = rectangle.Location.Y; i < rectangle.Location.Y + rectangle.Height; i += step)
            {
                points.Add(new Point(rectangle.Location.X, i));
                points.Add(new Point(rectangle.Location.X + rectangle.Width, i));
            }
            return points;
        }

        /// <summary>
        /// Returns a collection of rectangle's points around its circumference with specified step between them starting from middle out.
        /// If you want to return all points, set step to 1. 
        /// </summary>
        /// <param name="rectangle">The rectangle.</param>
        /// <param name="position">You will get only points from position which is specified in this parameter.</param>
        /// <param name="stepValue"></param>
        /// <returns></returns>
        public static IEnumerable<Point> GetPoints(this Rectangle rectangle, RelativePosition position, int? stepValue = null)
        {
            int step = Settings.Default.PathFindingSensitivity;

            if (stepValue.HasValue)
                step = stepValue.Value;

            var points = new List<Point>();

            switch (position)
            {
                case RelativePosition.OnTheLeft:
                    int leftStep = rectangle.Height/5;
                    for (int i = rectangle.Location.Y; i < rectangle.Location.Y + rectangle.Height; i += step)
                    {
                        if(i < (rectangle.Location.Y + (leftStep /2)) || i > (rectangle.Location.Y + (leftStep*17))) continue;
                        points.Add(new Point(rectangle.Location.X + rectangle.Width, i));
                    }
                    //if no points has been added, add one point in the middle
                    if(points.Count == 0) points.Add(new Point(rectangle.Location.X + rectangle.Width,rectangle.Location.Y + (rectangle.Height/2)));
                    return points;
                case RelativePosition.OnTheRight:
                    int rightStep = rectangle.Height /5;
                    for (int i = rectangle.Location.Y; i < rectangle.Location.Y + rectangle.Height; i += step)
                    {
                        if (i < (rectangle.Location.Y + (rightStep /2)) || i > (rectangle.Location.Y + (rightStep *17))) continue;
                        points.Add(new Point(rectangle.Location.X, i));
                    }
                    //if no points has been added, add one point in the middle
                    if(points.Count == 0) points.Add(new Point(rectangle.Location.X, rectangle.Location.Y + (rectangle.Height/2)));
                    return points;
                case RelativePosition.Above:
                    int topStep = rectangle.Width/10;
                    for (int i = rectangle.Location.X; i < rectangle.Location.X + rectangle.Width; i += step)
                    {
                        if (i < (rectangle.Location.X + (topStep)) || i > (rectangle.Location.X + (topStep * 22))) continue;
                        points.Add(new Point(i, rectangle.Location.Y));
                    }
                    //if no points has been added, add one point in the middle
                    if(points.Count == 0) points.Add(new Point(rectangle.Location.X + (rectangle.Width /2),rectangle.Location.Y));
                    return points;
                case RelativePosition.Below:
                    int bottomStep = rectangle.Width/10;
                    for (int i = rectangle.Location.X; i < rectangle.Location.X + rectangle.Width; i += step)
                    {
                        if (i < (rectangle.Location.X + (bottomStep)) || i > (rectangle.Location.X + (bottomStep* 22))) continue;
                        points.Add(new Point(i, rectangle.Location.Y + rectangle.Height));
                    }
                    //if no points has been added, add one point in the middle
                    if (points.Count == 0) points.Add(new Point(rectangle.Location.X + (rectangle.Width / 2), rectangle.Location.Y + rectangle.Height));
                    return points;
                case RelativePosition.Away:
                    return GetPoints(rectangle, stepValue);
                default:
                    throw new ArgumentOutOfRangeException("position");
            }
        }

        /// <summary>
        /// Reverts the text.
        /// </summary>
        /// <param name="source">The source.</param>
        /// <returns>Reverted string</returns>
        public static string RevertText(this string source)
        {
            var array = source.ToCharArray();
            Array.Reverse(array);
            return new string(array);
        }

        /// <summary>
        /// Calculates a relative position of the Rectangle to another Rectangle.
        /// </summary>
        /// <param name="rectangle">The rectangle.</param>
        /// <param name="other">The other.</param>
        /// <returns></returns>
        public static RelativePosition CalculatePositionTo(this Rectangle rectangle, Rectangle other)
        {
            if(new Rectangle(int.MinValue + other.X + 1,other.Y,int.MaxValue,other.Height).IntersectsWith(rectangle))
                return RelativePosition.OnTheLeft;

            if(new Rectangle(other.X,other.Y,int.MaxValue - other.X - 1,other.Height).IntersectsWith(rectangle))
                return RelativePosition.OnTheRight;

            if(new Rectangle(other.X,int.MinValue + other.Y -1,other.Width,int.MaxValue).IntersectsWith(rectangle))
                return RelativePosition.Below;

            if(new Rectangle(other.X,other.Y,other.Width,int.MaxValue - other.Y - 1).IntersectsWith(rectangle))
                return RelativePosition.Above;

            return RelativePosition.Away;
        }

        public static bool IsAbove(this Rectangle source, Rectangle other)
        {
            //var topRectangle = new Rectangle(int.MinValue/2,int.MinValue/2,int.MaxValue,int.MaxValue);
            //if (topRectangle.IntersectsWith(other))
            //    return true;
            //return false;

            return source.Bottom < other.Bottom;
        }

        /// <summary>
        /// Calculates the ange in degrees of <param name="line"></param> to a horizontal line.
        /// </summary>
        /// <param name="line">The line.</param>
        /// <returns>The degree value</returns>
        public static double CalculateAngeInDegrees(this Line line)
        {
            return Math.Atan2(line.Y1 - line.Y2, line.X1 - line.X2)*(180/Math.PI);
        }

        /// <summary>
        /// Gets the angle in degrees.
        /// </summary>
        /// <param name="line">The line.</param>
        /// <returns></returns>
        public static double GetAngleInDegrees(this Line line)
        {
            return Math.Atan2(line.Y1 - line.Y2, line.X1 - line.X2)*(180/Math.PI);
        }


        /// <summary>
        /// Gets string representation of give <see cref="MultiplicityType"/>.
        /// </summary>
        /// <param name="type">The type.</param>
        /// <returns></returns>
        public static string GetString(this MultiplicityType type)
        {
            switch (type)
            {
                case MultiplicityType.One:
                    return "1";
                case MultiplicityType.ZeroToOne:
                    return "0..1";
                case MultiplicityType.ZeroToMany:
                    return "0..*";
                case MultiplicityType.OneToMany:
                    return "1..*";
                default:
                    return string.Empty;
            }
        }

        /// <summary>
        /// Removes items from ObservableCollection by specified condition.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="coll">The coll.</param>
        /// <param name="condition">The condition.</param>
        /// <returns></returns>
        public static int Remove<T>(this ObservableCollection<T> coll, Func<T, bool> condition)
        {
            var itemsToRemove = coll.Where(condition).ToList();

            foreach (var itemToRemove in itemsToRemove)
            {
                coll.Remove(itemToRemove);
            }

            return itemsToRemove.Count;
        }

        /// <summary>
        /// Gets the closest point from collection of Points to specified Point.
        /// </summary>
        /// <param name="points">The points.</param>
        /// <param name="distanceTo">The distance to.</param>
        /// <returns>Point with closest distance.</returns>
        public static Point GetClosestPoint(this ICollection<Point> points, Point distanceTo)
        {
			if(points == null || ! points.Any()) throw new ArgumentException("points is null or empty");

            Point closestPoint = points.First();
            double closestDistance = double.MaxValue;

            foreach (var point in points)
            {
                var tempDistance = point.CalculateDistance(distanceTo);
                if (tempDistance < closestDistance)
                {
                    closestPoint = point;
                    closestDistance = tempDistance;
                }
            }

            return closestPoint;
        }
    }
}
