﻿using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using UmlDiagramDesigner.Core;
using UmlDiagramDesigner.Core.Enums;
using UmlDiagramDesigner.Core.Graph;
using UmlDiagramDesigner.DTO;
using UmlDiagramDesigner.Helpers;
using UmlDiagramDesigner.Models.PathFinders;
using UmlDiagramDesigner.ViewModels;
using UmlDiagramDesigner.ViewModels.DiagramItems;
using Point = System.Windows.Point;

namespace UmlDiagramDesigner.Models.Connectors
{
    public class TypeConnector : IAssociationConnector<DiagramItemViewModelBase>
    {
        public TypeConnector(IPathFinder pathFinder)
        {
            PathFinder = pathFinder;
            ShortestPathPathFinder = new ShortestPathPathFinder();
        }

        /// <summary>
        /// Gets or sets the path finder.
        /// </summary>
        /// <value>
        /// The path finder.
        /// </value>
        public IPathFinder PathFinder { get; set; }

        /// <summary>
        /// Gets or sets the shortest path path finder.
        /// </summary>
        /// <value>
        /// The shortest path path finder.
        /// </value>
        /// <remarks>This path finder is used for associations, that cannot be connected together with their type.</remarks>
        public PathFinders.IPathFinder ShortestPathPathFinder { get; set; }

        public void Connect(ICollection<IAssociation<DiagramItemViewModelBase>> associations)
        {
	        if (MainPageViewModel.IsImportingData) return;
            //get group of associations by type. each group will be connected together
            var groupsByType = GroupAssociationsByType(associations).ToArray();

            ResetAllPointProviders(groupsByType);

            foreach (var byTypeGroup in groupsByType)
            {
                //get the first association, all other association are going to connect to this one
                var first = byTypeGroup.Associations.First();

                //connect all association, that cannot be connected to the first one
                if (!first.IsConnectable && first.CanBend)
                {
                    foreach (var association in byTypeGroup.Associations)
                    {
                        ShortestPathPathFinder.SetPath(
                            ShortestPathPathFinder.FindPath(association.Source, association.Target), association);
                    }
                }
                else
                {
                    Point[] path = GetPath(first);

                    //first association is never connected to another
                    first.IsConnected = false;

                    PathFinder.SetPath(path, first);

                    var availableAssociations = new List<IAssociation<DiagramItemViewModelBase>>() {first};

                    //attach all other associations of the same type to the closest available path of the same type
                    foreach (var association in byTypeGroup.Associations.Where(a => !a.Equals(first)))
                    {
                        if (!association.IsConnectable)
                        {
                            //association cannot be connected to other associations, has to be straight (or with right angle's)
                            PathFinder.SetPath(GetPath(association),association);
                        }
                        else
                        {
                            //attach current association to one of the available ones
                            Attach(association, availableAssociations);
                            //this association is now available for other associations to connect to
                            availableAssociations.Add(association);
                        }
                    }
                }
            }
        }

        private void ResetAllPointProviders(IEnumerable<AssociationsByTypeDTO<DiagramItemViewModelBase>> groupsByType)
        {
            foreach (var byType in groupsByType)
            {
                foreach (var association in byType.Associations)
                {
                    association.Source.PointDistributor.Reset();
                    association.Target.PointDistributor.Reset();
                }
            }
        }

        private IEnumerable<AssociationsByTypeDTO<DiagramItemViewModelBase>> GroupAssociationsByType(
            IEnumerable<IAssociation<DiagramItemViewModelBase>> associations)
        {
            //group associations by their target (only those with the same target can be connected together)
            var groupByTarget = (from a in associations
                group a by a.Target
                into g
                select new {Target = g.Key, Associations = g}).ToArray();

            var groupsByType = new List<AssociationsByTypeDTO<DiagramItemViewModelBase>>();

            //loop through associations with the same target
            //group them by their type, so they can be connected together
            foreach (var byTarget in groupByTarget)
            {
                groupsByType.AddRange(from a in byTarget.Associations
                    group a by a.GetType()
                    into g
                    select new AssociationsByTypeDTO<DiagramItemViewModelBase>()
                    {
                        Type = g.Key,
                        Associations = g
                    });
            }

            return groupsByType;
        }

        /// <summary>
        /// Attaches the <param name="association"/> to <param name="targetAssociation"/>.
        /// </summary>
        /// <param name="association">The association.</param>
        /// <param name="targetAssociation">The target association.</param>
        private void Attach(IAssociation<DiagramItemViewModelBase> association,
            IAssociation<DiagramItemViewModelBase> targetAssociation)
        {
            //if path we are trying to connect to has a right angle, connect to it, if it doesn't find another point to connect to
            if (targetAssociation.AdditionalPoints != null && targetAssociation.AdditionalPoints.Any())
            {
                var path = GetPath(association, targetAssociation.AdditionalPoints.Last());
                PathFinder.SetPath(path, association);
            }
            else
            {
                var sinkPoint = new Point(
                    (targetAssociation.SourcePoint.X + targetAssociation.TargetPoint.X)/2,
                    (targetAssociation.SourcePoint.Y + targetAssociation.TargetPoint.Y)/2);

                var path = GetPath(association, sinkPoint);
                PathFinder.SetPath(path,association);
            }

            //set IsConnected to true, so end shape will not drawn
            association.IsConnected = true;
        }

        private void Attach(IAssociation<DiagramItemViewModelBase> association,
            IEnumerable<IAssociation<DiagramItemViewModelBase>> targetAssociations)
        {
            var availablePoints = new List<Point>();
            foreach (var availableAssociation in targetAssociations)
            {
                //if there are right angles in path, add them as available sink points
                if (availableAssociation.AdditionalPoints != null && availableAssociation.AdditionalPoints.Any())
                {
                    availablePoints.AddRange(availableAssociation.AdditionalPoints);
                }
                else
                {
                    //if not, get the middle of the path as sink point
                    availablePoints.Add(
                        new Point((availableAssociation.SourcePoint.X + availableAssociation.TargetPoint.X)/2,
                            (availableAssociation.SourcePoint.Y + availableAssociation.TargetPoint.Y)/2));
                }
            }

            var path = GetPath(association, availablePoints.ToArray());
            PathFinder.SetPath(path,association);

            association.IsConnected = true;
        }

        private Point[] GetPath(IAssociation<DiagramItemViewModelBase> first)
        {
            Point[] path;

            //if items are away from each other, straight line is not possible
            if (first.Source.Rectangle.CalculatePositionTo(first.Target.Rectangle) == RelativePosition.Away)
                path = PathFinder.FindPath(first.Source, first.Target, false);
            else
                path = PathFinder.FindPath(first.Source, first.Target, true);
            return path;
        }

        private Point[] GetPath(IAssociation<DiagramItemViewModelBase> first, Point targetPoint)
        {
            Point[] path;
            //if items are away from each other, straight line is not possible
            if (
                first.Source.Rectangle.CalculatePositionTo(
                new Rectangle((int)targetPoint.X, (int)targetPoint.Y, 1, 1)) == RelativePosition.Away)
                path = PathFinder.FindPath(first.Source, targetPoint, false);
            else
                path = PathFinder.FindPath(first.Source, targetPoint, true);
            return path;
        }

        /// <summary>
        /// Gets the path from <param name="association"/> to one of the <param name="points"/>.
        /// </summary>
        /// <param name="association">The association.</param>
        /// <param name="points">The points.</param>
        /// <returns>Shortest found path</returns>
        private Point[] GetPath(IAssociation<DiagramItemViewModelBase> association, Point[] points)
        {
            return PathFinder.FindPath(association.Source, points);
        }
    }
}
