﻿using System;
using System.Windows;
using UmlDiagramDesigner.Core;
using UmlDiagramDesigner.Models.Attributes;
using UmlDiagramDesigner.ViewModels;
using UmlDiagramDesigner.ViewModels.DiagramItems;
using UmlDiagramDesigner.ViewModels.DiagramRelationships;

namespace UmlDiagramDesigner.Models
{
    /// <summary>
    /// Creates concrete diagram items.
    /// </summary>
    public class DiagramItemFactory
    {
        private readonly IDialogService _dialogService;

        public DiagramItemFactory(IDialogService dialogService)
        {
            _dialogService = dialogService;
        }


        /// <summary>
        /// Creates a concrete diagram item with specific position on the canvas from <value>ToolboxItemViewModel</value>.
        /// </summary>
        /// <param name="item">Toolbox item, which was dragged on the canvas.</param>
        /// <param name="canvasPosition">The canvas position.</param>
        /// <returns></returns>
        public DiagramItemViewModelBase CreateItem(ToolboxItemViewModel item, Point canvasPosition)
        {
            try
            {
                var constructor = item.DiagramItemType.GetConstructor(Type.EmptyTypes);
                if (constructor == null) return null;

                var tempItem = (DiagramItemViewModelBase)constructor.Invoke(null);

                tempItem.LeftPosition = canvasPosition.X - (item.DefaultWidth / 2);
                tempItem.TopPosition = canvasPosition.Y;
                tempItem.Height = item.DefaultHeight;
                tempItem.Width = item.DefaultWidth;
                return tempItem;
            }
            catch (Exception e)
            {
                _dialogService.Display(e,string.Format("Error occured while creating diagram item."));
                return null;
            }
        }

        public RelationshipViewModelBase CreateRelationship(Type relationshipType, DiagramItemViewModelBase source,
            DiagramItemViewModelBase target)
        {
            try
            {
                if (!Attribute.IsDefined(source.GetType(), typeof(AssociatableAttribute)) || 
                    !Attribute.IsDefined(source.GetType(), typeof(AssociatableAttribute)))
                {
                    _dialogService.Display(string.Format("The target {0} is not associatable with other items.",source));
                    return null;
                }

                var constructor = relationshipType.GetConstructor(Type.EmptyTypes);
                if (constructor == null) return null;

                var relationship = (RelationshipViewModelBase)constructor.Invoke(null);

                relationship.Source = source;
                relationship.Target = target;

                source.Relationships.Add(relationship);
                target.Relationships.Add(relationship);

                return relationship;
            }
            catch (Exception e)
            {
                _dialogService.Display(e, string.Format("Error occured while creating diagram relationship."));
                return null;
            }
        }

        public RelationshipViewModelBase CreateRelationship(Type relationshipType, DiagramItemViewModelBase source,
        Point target)
        {
            try
            {
                if (source == null)
                {
                    _dialogService.Display(string.Format("Relationship has to be dropped on an item."));
                    return null;
                }
                if (!Attribute.IsDefined(source.GetType(),typeof(AssociatableAttribute)))
                {
                    _dialogService.Display(string.Format("The target is not associatable with other items."));
                    return null;
                }

                var constructor = relationshipType.GetConstructor(Type.EmptyTypes);
                if (constructor == null) return null;

                var tempRelationship = (RelationshipViewModelBase)constructor.Invoke(null);

                tempRelationship.Source = source;
                tempRelationship.TargetPoint = target;

                return tempRelationship;
            }
            catch (Exception e)
            {
                _dialogService.Display(e, string.Format("Error occured while creating diagram relationship."));
                return null;
            }
        }

        public RelationshipViewModelBase CreateRelationship(TemporaryRelationship relationship, DiagramItemViewModelBase target)
        {
            if (!Attribute.IsDefined(target.GetType(), typeof(AssociatableAttribute)))
            {
                _dialogService.Display("The target is not associatable with other items.");
                return null;
            }

            relationship.BaseRelationship.Source = relationship.Source;
            relationship.BaseRelationship.Target = target;
            relationship.BaseRelationship.Source.Relationships.Add(relationship.BaseRelationship);
            relationship.BaseRelationship.Target.Relationships.Add(relationship.BaseRelationship);
            return relationship.BaseRelationship;
        }
    }
}
