﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
using UmlDiagramDesigner.Core.Enums;
using UmlDiagramDesigner.Helpers;
using UmlDiagramDesigner.Properties;

namespace UmlDiagramDesigner.Models.PathFinders
{
    /// <summary>
    /// IPathFinder implementation, that finds path between two sets of points.
    /// It first tries to find a straight vertical or horizontal line, if it fails,
    /// it creates a right angle in the middle of first found the path.
    /// </summary>
    public class OrthogonalPathFinder : PathFinderBase
    {        
        private readonly IPathFinder _shortestPathFinder = new ShortestPathPathFinder();
        private double _anglePosition = 66;
        private int _tolerance = Settings.Default.PathFindingSensitivity;
        private int _effort = Settings.Default.PathFindingEffort;

        #region Properties

        /// <summary>
        /// Gets or sets the angle position. The number is a percentage, so the value has to be within 1-100 from target.
        /// Default value is 50.
        /// </summary>
        /// <value>
        /// The angle position.
        /// </value>
        public double AnglePosition
        {
            get { return _anglePosition; }
            set
            {
                if (value < 1 || value > 99) return;
                _anglePosition = value;
            }
        }

        /// <summary>
        /// Gets or sets the tolerance of comparing X or Y coordinates of points. 
        /// If set to 0, there is no tolerance.
        /// </summary>
        /// <value>
        /// The tolerance.
        /// </value>
        public int Tolerance
        {
            get { return _tolerance; }
            set
            {
                if (value < 0) return;
                _tolerance = value;
            }
        }

        /// <summary>
        /// Gets or sets the effort. This value represents a maximum number of attempts
        /// that this algorithm will use to find the most efficient path. Lower this number to increase performance.
        /// </summary>
        /// <value>
        /// The effort.
        /// </value>
        public int Effort
        {
            get { return _effort; }
            set { _effort = value; }
        }

        #endregion


        #region Private Methods

        private Point[] FindClosestPathWithRightAngle(Point[] setA, Point[] setB)
        {
            var shortestPath = _shortestPathFinder.FindPath(setA, setB);

            var finalPath = new List<Point>() {shortestPath[0], shortestPath[1]};

            AddRightAngle(finalPath);

            return finalPath.ToArray();
        }

        /// <summary>
        /// Adds the right angle. Path must contain at least two elements.
        /// </summary>
        /// <param name="straightPath">The straight path.</param>
        private void AddRightAngle(IList<Point> straightPath)
        {
            
            if(straightPath.Count < 1)
                throw new ArgumentException("straightPath");

            if (straightPath.Count > 2) //more points, than start and end, remove them
            {
                for (int i = 1; i < straightPath.Count - 1; i++)
                    straightPath.RemoveAt(i);
            }

            //straightPath.Insert(1,new Point(straightPath[1].X, straightPath[0].Y));

            switch (SourceToTargetPosition)
            {
                case RelativePosition.OnTheLeft:
                    straightPath.Insert(1, new Point((straightPath[0].X + straightPath[1].X)/2, straightPath[0].Y));
                    straightPath.Insert(2, new Point(straightPath[1].X, straightPath[2].Y));
                    break;
                case RelativePosition.OnTheRight:
                    straightPath.Insert(1, new Point((straightPath[0].X + straightPath[1].X)/2, straightPath[0].Y));
                    straightPath.Insert(2, new Point(straightPath[1].X, straightPath[2].Y));
                    break;
                case RelativePosition.Above:
                    straightPath.Insert(1, new Point(straightPath[0].X, (straightPath[0].Y + straightPath[1].Y)/2));
                    straightPath.Insert(2, new Point(straightPath[2].X, straightPath[1].Y));
                    break;
                case RelativePosition.Below:
                    straightPath.Insert(1, new Point(straightPath[0].X, (straightPath[0].Y + straightPath[1].Y) / 2));
                    straightPath.Insert(2, new Point(straightPath[2].X, straightPath[1].Y));
                    break;
                case RelativePosition.Away:
                    straightPath.Insert(1, new Point(straightPath[0].X, (straightPath[0].Y + straightPath[1].Y) / 2));
                    straightPath.Insert(2, new Point(straightPath[1].X, straightPath[2].Y));
                    break;
                case null:
                    break;
                default:
                    throw new ArgumentOutOfRangeException();
            }
        }

        private Tuple<Point, Point> FindClosestStraightTuple(Point[] setA, Point[] setB, bool horizontal)
        {
            var tuples = GetStraightTuples(setA, setB, horizontal);

            if (!tuples.Any()) //no horizontal or vertical tuple was found
                return null;

            //group them by first point
            var groupsByPoint = (from point in tuples
                group point by point.Item1
                into g
                select new {Point = g.Key, Neighbours = g}).ToArray();

            var closestDistance = double.MaxValue;
            Tuple<Point, Point> closestTuple = null;

            foreach (var groupByPoint in groupsByPoint)
            {
                foreach (var neighbour in groupByPoint.Neighbours)
                {
                    //calculate their distance, if it is lower, than current closest, set it as closest
                    var tempDistance = neighbour.Item1.CalculateDistance(neighbour.Item2);
                    if (tempDistance < closestDistance)
                    {
                        closestDistance = tempDistance;
                        closestTuple = new Tuple<Point, Point>(neighbour.Item1, neighbour.Item2);
                    }
                }
            }

            //return the shortest path
            return closestTuple;
        }

        /// <summary>
        /// Gets the straight tuples. Searches for matching X or Y coordinates, depending on the value of <c>horizontal</c> parameter.
        /// </summary>
        /// <param name="setA">The set a.</param>
        /// <param name="setB">The set b.</param>
        /// <param name="horizontal">if set to <c>true</c> [horizontal].</param>
        /// <returns>Collection of Tuple's of two Points, where their X or Y coordinates are the same.</returns>
        /// <remarks>Tolerance value from settings is taken into consideration when comparing coordinates.</remarks>
        private ICollection<Tuple<Point, Point>> GetStraightTuples(Point[] setA, Point[] setB, bool horizontal)
        {
            var tuples = new List<Tuple<Point, Point>>();

            foreach (var pointA in setA)
            {
                //find all points that have the same X or Y value and add them into a collection
                Point a = pointA;
                tuples.AddRange(
                    setB.Where(pointB =>
                    {
                        if (!horizontal)
                            //if their X or Y value is the same, they are horizontaly or verticaly in a line with each other
                            return Math.Abs(pointB.X - a.X) < Tolerance;
                        return Math.Abs(pointB.Y - a.Y) < Tolerance;
                    })
                        .Select(pointB => new Tuple<Point, Point>(pointA, pointB)));
            }

            return tuples;
        }

        private Point[] FindClosestStraightPath(Point[] setA, Point[] setB)
        {
            var horizontal = FindClosestStraightTuple(setA, setB, true);
            if (horizontal != null) // vertical line was not found, return horizontal
                return new[] { horizontal.Item1, horizontal.Item2 };

            var vertical = FindClosestStraightTuple(setA, setB, false);
            if (vertical != null) // horizontal line was not found, return vertical
                return new[] { vertical.Item1, vertical.Item2 };

           return FindClosestPathWithRightAngle(setA, setB);
        }

        #endregion


        #region Public methods

        public override Point[] FindPath(Point[] setA, Point[] setB)
        {
            var horizontal = FindClosestStraightTuple(setA, setB, true);
            var vertical = FindClosestStraightTuple(setA, setB, false);

            if (horizontal == null && vertical == null)
                return FindClosestPathWithRightAngle(setA, setB);

            if (horizontal == null) // horizontal line was not found, return vertical
                return new[] { vertical.Item1, vertical.Item2 };

            return new[] { horizontal.Item1, horizontal.Item2 };
        }

        /// <summary>
        /// Faster implementation, which based on value of <c>straightLinePossible</c> parameter
        /// looks only for paths right angle or for all possible paths, like the original implementation.
        /// If <c>straightLinePossible</c> is set to false this method automatically searches for paths with right angle.
        /// </summary>
        /// <param name="setA">The set a.</param>
        /// <param name="setB">The set b.</param>
        /// <param name="straightLinePossible"></param>
        /// <returns></returns>
        public override Point[] FindPath(Point[] setA, Point[] setB, bool straightLinePossible)
        {
            if (!straightLinePossible)
                //if straight path is not possible we can skip the search for it and find one with right angle
            {
                return FindClosestPathWithRightAngle(setA, setB);
            }
            return FindClosestStraightPath(setA, setB);
        }

        #endregion
    }
}
