﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using UmlDiagramDesigner.Core.Enums;
using UmlDiagramDesigner.Core.Graph;
using UmlDiagramDesigner.Helpers;
using UmlDiagramDesigner.Properties;
using UmlDiagramDesigner.ViewModels.DiagramItems;
using Point = System.Windows.Point;

namespace UmlDiagramDesigner.Models.PathFinders
{
    public abstract class PathFinderBase : IPathFinder
    {
        protected RelativePosition? SourceToTargetPosition;

        public Point[] FindPath(DiagramItemViewModelBase item1, DiagramItemViewModelBase item2)
        {
            var item1PointPosition = GetPointPosition(item1.Rectangle, item2.Rectangle);
            var item2PointPosition = GetPointPosition(item2.Rectangle, item1.Rectangle);

            SourceToTargetPosition = item1PointPosition;

            return FindPath(item1.GetAvailablePoints(item1PointPosition), item2.GetAvailablePoints(item2PointPosition));
        }

        public Point[] FindPath(DiagramItemViewModelBase item1, DiagramItemViewModelBase item2,
            bool possibleStraightLine)
        {
            var item1PointPosition = GetPointPosition(item1.Rectangle, item2.Rectangle);
            var item2PointPosition = GetPointPosition(item2.Rectangle, item1.Rectangle);

            SourceToTargetPosition = item1PointPosition;

            return FindPath(item1.GetAvailablePoints(item1PointPosition), item2.GetAvailablePoints(item2PointPosition), possibleStraightLine);
            //return FindPath(
            //    new[] {item1.PointDistributor.GetPoint(item1PointPosition)},
            //    new[] {item2.PointDistributor.GetPoint(item2PointPosition)}, 
            //    possibleStraightLine);
        }

        public Point[] FindPath(DiagramItemViewModelBase item, Point point, bool possibleStraightLine)
        {
            var itemPointPosition = GetPointPosition(item.Rectangle, new Rectangle((int) point.X, (int) point.Y, 1, 1));

            SourceToTargetPosition = itemPointPosition;

            return FindPath(item.GetAvailablePoints(itemPointPosition), new[] {point}, possibleStraightLine);
        }

        public Point[] FindPath(DiagramItemViewModelBase item, Point[] points)
        {
            var paths = new List<Point[]>();

            //calculate paths for all available points
            foreach (var point in points)
            {
                //get position for particular point so we know if there is straight line possible
                var position = item.Rectangle.CalculatePositionTo(new Rectangle((int) point.X, (int) point.Y, 1, 1));

                if (position == RelativePosition.Away)
                    paths.Add(FindPath(item, point, false));
                else
                    paths.Add(FindPath(item,point,true));
            }

            //get first path as default
            Point[] shortest = paths.FirstOrDefault();
            double shortestDistance = double.MaxValue;

            //determine which path is shortest
            foreach (var path in paths)
            {
                double pathDistance = 0;
                //calculate distance for whole path
                for (int i = 0; i < path.Length; i++)
                {
                    //check if there is another point if path
                    if(i == path.Length - 1) continue;
                    //add distance to pathDistance variable
                    pathDistance += path[i].CalculateDistance(path[i + 1]);
                }
                //if current path is shorter than the previous, select it as shorterst
                if (pathDistance < shortestDistance)
                {
                    shortestDistance = pathDistance;
                    shortest = path;
                }
            }

            return shortest;
        }

        public Point[] FindPath(Rectangle rect1, Rectangle rect2)
        {
            var rect1PointPosition = GetPointPosition(rect1, rect2);
            var rect2PointPosition = GetPointPosition(rect2, rect1);

            SourceToTargetPosition = rect1PointPosition;

            return FindPath(rect1.GetPoints(rect1PointPosition, Settings.Default.PathFindingSensitivity).ToArray(),
                rect2.GetPoints(rect2PointPosition, Settings.Default.PathFindingSensitivity).ToArray());
        }

        public Point[] FindPath(Rectangle rect1, Rectangle rect2, bool possibleStraightLine)
        {
            var rect1PointPosition = GetPointPosition(rect1,rect2);
            var rect2PointPosition = GetPointPosition(rect2,rect1);

            SourceToTargetPosition = rect1PointPosition;

            return FindPath(rect1.GetPoints(rect1PointPosition, Settings.Default.PathFindingSensitivity).ToArray(),
                rect2.GetPoints(rect2PointPosition, Settings.Default.PathFindingSensitivity).ToArray(),possibleStraightLine);
        }

        public abstract Point[] FindPath(Point[] setA, Point[] setB, bool possibleStraightLine);

        public abstract Point[] FindPath(Point[] setA, Point[] setB);

        public void SetPath(Point[] points, IAssociation<DiagramItemViewModelBase> association)
        {
            if (points == null || points.Length < 2) return; //not proper path data

            association.SourcePoint = points[0];
            association.TargetPoint= points[points.Length - 1];

            association.AdditionalPoints.Clear();
            for (int i = 1; i < points.Length - 1; i++)
                association.AdditionalPoints.Add(points[i]);
        }

        /// <summary>
        /// Gets the position from which points will be collected (if rectangle is above another rectangle, we need to get 
        /// points from it's bottom side) Bottom -> Above, OnTheLeft -> OnTheRight and so on.
        /// </summary>
        /// <returns></returns>
        /// <exception cref="System.ArgumentOutOfRangeException">relativePosition</exception>
        private RelativePosition GetPointPosition(Rectangle rect1, Rectangle rect2)
        {
            var position = rect1.CalculatePositionTo(rect2);

            switch (position)
            {
                case RelativePosition.OnTheLeft:
                    return RelativePosition.OnTheLeft;
                case RelativePosition.OnTheRight:
                    return RelativePosition.OnTheRight;
                case RelativePosition.Above:
                    return RelativePosition.Above;
                case RelativePosition.Below:
                    return RelativePosition.Below;
                case RelativePosition.Away:
                    if (rect1.IsAbove(rect2))
                        return RelativePosition.Below;
                    return RelativePosition.Above;
                default:
                    throw new ArgumentOutOfRangeException();
            }
        }
    }
}
