﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.Drawing;
using System.Linq;
using System.Windows;
using System.Windows.Input;
using Newtonsoft.Json;
using UmlDiagramDesigner.Core;
using UmlDiagramDesigner.Core.Graph;
using UmlDiagramDesigner.DAL;
using UmlDiagramDesigner.DTO;
using UmlDiagramDesigner.Helpers;
using UmlDiagramDesigner.Models;
using UmlDiagramDesigner.Models.Connectors;
using UmlDiagramDesigner.Models.DiagramValidation;
using UmlDiagramDesigner.Models.PathFinders;
using UmlDiagramDesigner.Services;
using UmlDiagramDesigner.ViewModels.Commands;
using UmlDiagramDesigner.ViewModels.DiagramItems;
using UmlDiagramDesigner.ViewModels.DiagramRelationships;
using Point = System.Windows.Point;

namespace UmlDiagramDesigner.ViewModels
{
    public class CanvasViewModel : ViewModelBase, IEntity<string>, IIndexedGraph
    {
        #region Variables

        private ObservableCollection<DiagramEntityViewModelBase> _items;
        private RelationshipViewModelBase _tempRelationship;
        private DiagramItemFactory _itemFactory;
        private static int _width, _height;
        private Point _mousePosition;

        private IDialogService _dialogService;
        private IRelationshipVerifier _relationshipVerifier;
        private static CanvasViewModel _instance;

        #endregion

        #region Constructors

        public CanvasViewModel()
        {
            Initialize();
            _instance = this;
        }

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets the identifier.
		/// </summary>
		/// <value>
		/// The identifier.
		/// </value>
		public string Id { get; set; }

		/// <summary>
		/// Gets or sets the name.
		/// </summary>
		/// <value>
		/// The name.
		/// </value>
		public string Name { get; set; }

		/// <summary>
		/// Gets or sets the items. This collection contains all items that are presented on diagram including all relationships.
		/// </summary>
		/// <value>
		/// The items.
		/// </value>
		public ObservableCollection<DiagramEntityViewModelBase> Items
        {
            get { return _items; }
            set
            {
                
                _items = value;
                if (_items != null)
                {
                    _items.CollectionChanged += ItemsCollectionChanged;
                    foreach (var item in value)
                    {
                        var di = item as DiagramItemViewModelBase;
                        if (di == null) continue;
                        di.ItemMoved += DiagramItemMoved;
                    }
                }

                OnPropertyChanged();
            }
        }

		/// <summary>
		/// Gets a value indicating whether mouse is hovering over any item on diagram.
		/// </summary>
		/// <value>
		/// <c>true</c> if any item is hovered; otherwise, <c>false</c>.
		/// </value>
		[Newtonsoft.Json.JsonIgnore]
        public bool IsAnyItemHovered
        {
            get
            {
                var rectangles = new List<Rectangle>();
                foreach (var element in Items)
                {
                    var item = element as DiagramItemViewModelBase;
                    if (item != null)
                    {
                        var rectangle = item.Rectangle;
                        rectangle.Offset(5, 5);
                        rectangles.Add(rectangle);
                    }
                    else
                    {
                        if (((RelationshipViewModelBase) element).IsHovered)
                        {
                            return true;
                        }
                    }
                }
                return MousePosition.IsPartOfAny(rectangles);
            }
        }

		/// <summary>
		/// Gets the width of diagram canvas.
		/// </summary>
		/// <value>
		/// The width.
		/// </value>
		public static int Width { get { return _width; } }

		/// <summary>
		/// Gets the height of diagram canvas.
		/// </summary>
		/// <value>
		/// The height.
		/// </value>
		public static int Height { get { return _height; } }

		/// <summary>
		/// Gets or sets the mouse position.
		/// </summary>
		/// <value>
		/// The mouse position.
		/// </value>
		[Newtonsoft.Json.JsonIgnore]
        public Point MousePosition
        {
            get { return _mousePosition; }
            set
            {
                _mousePosition = value;
                if (MouseMoved != null)
                    MouseMoved(this, _mousePosition);
            }
        }

		/// <summary>
		/// Gets or sets the currently used connector engine.
		/// </summary>
		/// <value>
		/// The connector engine.
		/// </value>
		[Newtonsoft.Json.JsonIgnore]
        public IConnectorEngine ConnectorEngine { get; set; }

		/// <summary>
		/// Gets the instance.
		/// </summary>
		/// <value>
		/// The instance.
		/// </value>
		[Newtonsoft.Json.JsonIgnore]
        public static CanvasViewModel Instance { get { return _instance; } }

		#endregion

		#region Private Methods

		/// <summary>
		/// Initializes this instance.
		/// </summary>
		private void Initialize()
        {
            Id = Guid.NewGuid().ToString();
            Items = new ObservableCollection<DiagramEntityViewModelBase>();
            _itemFactory = new DiagramItemFactory(new WpfDialogService());
            _width = 0;
            _height = 0;
            _dialogService = new WpfDialogService();
            ConnectorEngine = new ConnectorEngine(new TypeConnector(new OrthogonalPathFinder()));
            _relationshipVerifier = new RelationshipVerifier(_dialogService);

            Messenger.Register<DeletedEntityDTO>(this, toDelete => DeleteEntity(toDelete.EntityToDelete));
            Messenger.Register<DeletedRelationshipDTO>(this, toDelete => DeleteRelationship(toDelete.Relationship));
        }

		/// <summary>
		/// Raises ZIndex property of selected diagram if it is colliding with any other item, so it is moved to foreground.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="eventArgs">The <see cref="EventArgs"/> instance containing the event data.</param>
		private void DiagramItemOnSelected(object sender, EventArgs eventArgs)
        {
            var selectedItem = sender as DiagramItemViewModelBase;
            if (selectedItem == null) return;

            var selectedItemRectangle = selectedItem.Rectangle;

            foreach (var diagramItem in Items.OfType<DiagramItemViewModelBase>().Where(item => !item.Equals(selectedItem)))
            {
                //checks if two diagram items collide with each other
                //if they do and selected item is in the background, then raise it's ZIndex to bring it to the foreground
                if (selectedItemRectangle.IntersectsWith(diagramItem.Rectangle))
                {
                    if (selectedItem.ZIndex <= diagramItem.ZIndex)
                    {
                        selectedItem.ZIndex += (diagramItem.ZIndex - selectedItem.ZIndex) + 1;
                    } 
                }
            }
        }



        #region Items Collection Changed

        private void ItemsCollectionChanged(object sender,NotifyCollectionChangedEventArgs e)
        {
            switch (e.Action)
            {
                case NotifyCollectionChangedAction.Add:
                    ItemAdded(e.NewItems);
                    break;
                case NotifyCollectionChangedAction.Remove:
                    ItemRemoved(e.OldItems);
                    break;
                case NotifyCollectionChangedAction.Reset:
                    ItemsResetted(e.OldItems);
                    break;
            }
        }

        private void ItemAdded(IList newItems)
        {
            if (newItems.Count == 1)
            {
                var item = newItems[0] as DiagramItemViewModelBase;

                if (item == null) return;

                item.ItemMoved += DiagramItemMoved;
            }
        }

        private void ItemRemoved(IList oldItems)
        {
            if (oldItems.Count == 1)
            {
                var relationship = oldItems[0] as RelationshipViewModelBase;
                if (relationship == null || relationship is TemporaryRelationship) return;

                relationship.Source.Relationships.Remove(relationship);
                relationship.Target.Relationships.Remove(relationship);
            }
        }

        private void ItemsResetted(IList oldItems)
        {
            if (oldItems == null) return; 

            foreach (var oldItem in oldItems)
            {
                var relationship = oldItem as RelationshipViewModelBase;
                if (relationship == null || relationship is TemporaryRelationship) return;

                relationship.Source.Relationships.Remove(relationship);
                relationship.Source.Relationships.Remove(relationship);
            }
        }

		#endregion


		/// <summary>
		/// Deselects all diagram items.
		/// </summary>
		private void DeselectAllDiagramItems()
        {
            foreach (var item in Items.OfType<DiagramItemViewModelBase>())
            {
                item.IsSelected = false;
            }
        }

		/// <summary>
		/// If no item is clicked on diagram, it deselects all items and removes TemporaryRelationship.
		/// If any item is clicked and temporary relationship is not null, it tries to create relationship, otherwise it selects the clicked item.
		/// </summary>
		/// <param name="clickPosition">The click position.</param>
		private void DiagramLeftButtonClicked(Point clickPosition)
        {
            if (!clickPosition.IsPartOfAny(_items.OfType<DiagramItemViewModelBase>().Select(item => item.Rectangle))) //no item was clicked
            {
                DeselectAllDiagramItems();
                MouseMoved -= CanvasOnMouseMoved;

                if (_tempRelationship != null)
                {
                    //clicked on empty spot while dragging temporary relationship
                    RemoveTempRelationship();
                }
            }
            else //something was a target of the click...
            {
                var clickedItem = _items.OfType<DiagramItemViewModelBase>().GetTopIntersectingItem(clickPosition); //get top most clicked item

                if (_tempRelationship != null && clickedItem != null && !clickedItem.Equals(_tempRelationship.Source)) //cannot create relationship with same Source and Target
                {
                    _tempRelationship.Target = clickedItem;
                    if(_relationshipVerifier.CanCreate(_tempRelationship))
                    {
                        CreateNewRelationship(clickedItem);
                    }
                    //no need for temp relationship anymore
                    RemoveTempRelationship();
                }
                else if(clickedItem != null)
                {
                    //item was clicked (not with selection box), so we have to select it and deselect all other selected items
                    DeselectAllDiagramItems();
                    clickedItem.IsSelected = true;
                }
            }
        }

		/// <summary>
		/// Creates new relationship.
		/// </summary>
		/// <param name="clickedItem">The clicked item.</param>
		private void CreateNewRelationship(DiagramItemViewModelBase clickedItem)
        {
            var createdRelationship =
                _itemFactory.CreateRelationship((TemporaryRelationship) _tempRelationship, clickedItem);

            //if it is foreign key relationship, user has to set name of the foreign key, so we need to call edit first
            if (createdRelationship is IForeignKeyRelationship)
            {
                createdRelationship.EditRelationshipCommand.Execute(null);
                //add created relationship to the diagram and remove the temporary one if edit was successful
                _items.Add(createdRelationship);
            }
            else
            {
                //no need to open edit window
                _items.Add(createdRelationship);
            }

            //new connection has been added, set its proper path
            ConnectorEngine.UpdateConnections(clickedItem);
        }

		/// <summary>
		/// Removes the temporary relationship.
		/// </summary>
		private void RemoveTempRelationship()
        {
            if (Items.Contains(_tempRelationship))
                Items.Remove(_tempRelationship);
            if (_tempRelationship.Source != null && _tempRelationship.Source.Relationships.Contains(_tempRelationship))
            {
                _tempRelationship.Source.Relationships.Remove(_tempRelationship);
            }
            if (_tempRelationship.Target != null && _tempRelationship.Target.Relationships.Contains(_tempRelationship))
            {
                _tempRelationship.Target.Relationships.Remove(_tempRelationship);
            }
            _tempRelationship = null;
        }

		/// <summary>
		/// Deletes the selected items.
		/// </summary>
		private void DeleteSelectedItems()
        {
            var itemsToRemove = new List<DiagramEntityViewModelBase>(); 
            foreach (var item in Items.OfType<DiagramItemViewModelBase>().Where(test => test.IsSelected))
            {
                itemsToRemove.Add(item);
                var relationshipsToRemove = Items.OfType<RelationshipViewModelBase>() //get all connections, that were part of the item, that should be removed
                    .Where(relationship => relationship.Target.Equals(item) || relationship.Source.Equals(item)).ToArray();

                itemsToRemove.AddRange(relationshipsToRemove);
            }
            foreach (var toRemove in itemsToRemove)
                Items.Remove(toRemove);
        }

		/// <summary>
		/// Deletes the entity.
		/// </summary>
		/// <param name="entity">The entity.</param>
		private void DeleteEntity(DiagramEntityViewModelBase entity)
        {
            if (entity == null) return;

            if (entity is DiagramItemViewModelBase)
            {
                var item = entity as DiagramItemViewModelBase;

                foreach (var relationship in item.Relationships.ToArray())
                    Items.Remove(relationship);

                Items.Remove(item);
            }
            else
            {
                Items.Remove(entity);
            }
        }
		/// <summary>
		/// Deletes the relationship.
		/// </summary>
		/// <param name="relationship">The relationship.</param>
		private void DeleteRelationship(RelationshipViewModelBase relationship)
        {
            if (Items.Contains(relationship))
            {
                Items.Remove(relationship);
            }
        }


		/// <summary>
		/// Selects the items in rectangle.
		/// </summary>
		/// <param name="rectangle">The rectangle.</param>
		private void SelectItemsInRectangle(Rect rectangle)
        {
            DeselectAllDiagramItems(); //unselect previously selected items (item's selected by click for example)

            foreach (var item in _items.OfType<DiagramItemViewModelBase>() //get all diagram items
                .Where(item => rectangle.Contains(new Rect(item.Rectangle.Location.X, item.Rectangle.Location.Y,
                    item.Rectangle.Width, item.Rectangle.Height))))//check if their position is within selected rectangle
            {
                item.IsSelected = true; //if they are, select them
            }
        }

		/// <summary>
		/// Updates connections when item has moved.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
		private void DiagramItemMoved(object sender, EventArgs e)
        {
            var item = sender as DiagramItemViewModelBase;

            if (item == null) return;

            ConnectorEngine.UpdateConnections(item);
        }

		/// <summary>
		/// If temporary relationship exists, its position is updated to provided <param name="point"></param>.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="point">The point.</param>
		private void CanvasOnMouseMoved(object sender, Point point)
        {
            if (_tempRelationship != null)
                _tempRelationship.TargetPoint = point;
        }

		/// <summary>
		/// Sets unique name to the item.
		/// </summary>
		/// <param name="item">The item.</param>
		private void SetUniqueName(DiagramItemViewModelBase item)
        {
            string name;
            int i = Items.OfType<DiagramItemViewModelBase>().Count();
            do
            {
                name = string.Format("{0}{1}", item.Name, i);
                i++;
            } while (Items.OfType<DiagramItemViewModelBase>().Any(it => it.Name.Equals(name)));

            item.Name = name;
        }

        #endregion

        #region Commands

        [Newtonsoft.Json.JsonIgnore]
        public GenericCommand<Point> DiagramCanvasMouseLeftButtonDownCommand
        {
            get { return new GenericCommand<Point>(DiagramLeftButtonClicked);}
        }

        [Newtonsoft.Json.JsonIgnore]
        public GenericCommand<Point> DiagramCanvasMouseMovedCommand
        {
            get { return new GenericCommand<Point>(point => MousePosition = point);}
        }

        [Newtonsoft.Json.JsonIgnore]
        public GenericCommand<Rect> SelectItemsInSelectionBoxCommand
        {
            get { return new GenericCommand<Rect>(SelectItemsInRectangle);}
        }

        [Newtonsoft.Json.JsonIgnore]
        public ICommand DeleteSelectedItemsCommand
        {
            get { return new GenericCommand(DeleteSelectedItems);}
        }

		#endregion

		#region Public Methods

		/// <summary>
		/// Creates the dropped item on canvas.
		/// </summary>
		/// <param name="item">The item.</param>
		/// <param name="droppedPosition">The dropped position.</param>
		public void CreateDroppedItemOnCanvas(ToolboxItemViewModel item, Point droppedPosition)
        {
            //create item
            if (item.DiagramItemType.IsSubclassOf(typeof (DiagramItemViewModelBase)))
            {
                var diagramItem = _itemFactory.CreateItem(item, droppedPosition);
                if (diagramItem == null) return;

                diagramItem.Selected += DiagramItemOnSelected;

                SetUniqueName(diagramItem);

                Items.Add(diagramItem);   
            }
            //create relationship
            else if (item.DiagramItemType.IsSubclassOf(typeof(RelationshipViewModelBase)))
            {
                var tempRelationship = _itemFactory.CreateRelationship(item.DiagramItemType,
                    _items.OfType<DiagramItemViewModelBase>().GetTopIntersectingItem(droppedPosition), droppedPosition);

                if (tempRelationship == null) return;
                _tempRelationship = new TemporaryRelationship(tempRelationship); //set decorator to temporary relationship, because of Opacity value
                MouseMoved += CanvasOnMouseMoved; //subscribe to an event for receiving changes of mouse position (which is then set to temp relationship target point)
                Items.Add(_tempRelationship); //temporarily add it to Items so it can be displayed on canvas
            }
        }

        public IEnumerable<IIndexedVertex> GetVertices()
        {
            return Items.OfType<DiagramItemViewModelBase>();
        }

        IEnumerable<IIndexedVertex> IIndexedGraph.GetVertices()
        {
            return Items.OfType<DiagramItemViewModelBase>();
        }

        #endregion

        private event EventHandler<Point> MouseMoved;
    }
}