﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.Linq;
using System.Windows.Input;
using UmlDiagramDesigner.Assets;
using UmlDiagramDesigner.Core;
using UmlDiagramDesigner.Core.Enums;
using UmlDiagramDesigner.Models.Attributes;
using UmlDiagramDesigner.ViewModels.Commands;
using UmlDiagramDesigner.ViewModels.DependentElements;
using UmlDiagramDesigner.ViewModels.DiagramRelationships;
using UmlDiagramDesigner.ViewModels.InputForms;
using UmlDiagramDesigner.Views.InputForms;

namespace UmlDiagramDesigner.ViewModels.DiagramItems
{
    [Associatable(SupportedAssocation  = UmlAssociation.Inheritance | UmlAssociation.Dependency |
        UmlAssociation.Aggregation | UmlAssociation.Composition | UmlAssociation.Association)]   
    public class ClassDiagramItemViewModel : DiagramItemViewModelBase
    {
        private ObservableCollection<AttributeViewModel> _attributes;
        private ClassEditViewModel _classEditViewModel;
        private ObservableCollection<TriggerViewModel> _triggers;

        public ClassDiagramItemViewModel()
        {
            Name = "ClassName";
            Attributes = new ObservableCollection<AttributeViewModel>();
            Triggers = new ObservableCollection<TriggerViewModel>();
        }

        #region Properties

        public ObservableCollection<AttributeViewModel> Attributes
        {
            get { return _attributes; }
            set
            {
                _attributes = value;
                OnPropertyChanged();
                if (_attributes != null)
                {
                    _attributes.CollectionChanged += AttributesOnCollectionChanged;
                }
            }
        }

        public ObservableCollection<TriggerViewModel> Triggers
        {
            get { return _triggers; }
            set { _triggers = value; OnPropertyChanged();}
        }

        #endregion

        #region Commands

        public ICommand OpenEditWindowCommand
        {
            get { return new GenericCommand(OpenEditWindow); }
        }

        #endregion

        #region Public methods

        public override string GetResourceStyleName()
        {
            return "EntityStyle";
        }

        #endregion

        /// <summary>
        /// Opens the edit window.
        /// </summary>
        private void OpenEditWindow()
        {
            var window = new ClassEditWindow() {DataContext = _classEditViewModel = new ClassEditViewModel(this)
            {
                Name = this.Name ?? string.Empty,
                //do not include foreign keys, they cannot be editted here, they should be editted in relationship edit window only
                Attributes = new ObservableCollection<AttributeViewModel>(Attributes.Where(att => !att.IsForeignKey)),
                Triggers = new ObservableCollection<TriggerViewModel>(Triggers)
            }};

            Name = _classEditViewModel.Name;

            _classEditViewModel.InputCompleted += (sender, args) =>
            {
                Name = _classEditViewModel.Name;
                //create new collection with editted values from edit windows plus all foreign keys
                Attributes = new ObservableCollection<AttributeViewModel>(
                    _classEditViewModel.Attributes
                    .Union(Attributes.Where(att => att.IsForeignKey))
                    .OrderByDescending(att => att)
                    .ToList());
                Triggers = _classEditViewModel.Triggers;

	            foreach (var removedAttribute in _classEditViewModel.RemovedAttributes)
	            {
		            RaiseAttributeRemovedEvent(removedAttribute);
	            }
            };

            window.ShowDialog();
        }

        private void AttributesOnCollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
        {
            if (e.Action == NotifyCollectionChangedAction.Remove)
            {
                foreach (var item in e.OldItems)
                {
                    var attribute = item as AttributeViewModel;
                    if (attribute != null)
                    {
                        RaiseAttributeRemovedEvent(attribute);
                    }
                }
            }
        }

        private void RaiseAttributeRemovedEvent(AttributeViewModel attribute)
        {
            if (AttributeRemoved != null)
                AttributeRemoved(this, attribute);
        }

        public event EventHandler<AttributeViewModel> AttributeRemoved;
    }
}
