﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Windows.Input;
using Newtonsoft.Json;
using UmlDiagramDesigner.Core;
using UmlDiagramDesigner.Core.Enums;
using UmlDiagramDesigner.Core.Graph;
using UmlDiagramDesigner.DTO;
using UmlDiagramDesigner.Helpers;
using UmlDiagramDesigner.Models;
using UmlDiagramDesigner.Models.PathFinders;
using UmlDiagramDesigner.Properties;
using UmlDiagramDesigner.ViewModels.Commands;
using UmlDiagramDesigner.ViewModels.DiagramRelationships;
using Point = System.Windows.Point;

namespace UmlDiagramDesigner.ViewModels.DiagramItems
{
    public abstract class DiagramItemViewModelBase : DiagramEntityViewModelBase, ISelectable, IIndexedVertex, IPointProvider
    {
        #region Variables

        private readonly string _id;
        private double _leftPosition, _topPosition, _width, _height;
        private readonly List<RelationshipViewModelBase> _relationships;
        private int _zIndex;
        private bool _isSelected;
        private Rectangle _rectangle;
        private bool _isExpanded = true;
        private double _actualHeight, _actualWidth;
        private readonly double _minHeight, _minWidth;

        private PointDistributor _pointDistributor;
        private string _name;

        #endregion

        #region Constructors

        protected DiagramItemViewModelBase()
        {
            _id = Guid.NewGuid().ToString();
            _relationships = new List<RelationshipViewModelBase>();
            this.Rectangle = CreateRectangle();
            _minWidth = 80;
        }

        protected DiagramItemViewModelBase(double topPosition, double leftPosition, double width, double height, double minWidth = 80, double minHeight = 120)
        {
            _id = Guid.NewGuid().ToString();
            _relationships = new List<RelationshipViewModelBase>();
            TopPosition = topPosition;
            LeftPosition = leftPosition;
            Width = width;
            Height = height;
            _minWidth = minWidth;
            _minHeight = minHeight;
            this.Rectangle = CreateRectangle();
        }

        #endregion

        #region Properties


        public string Name
        {
            get { return _name; }
            set
            {
                if (value == _name) return;
                _name = value;
                OnPropertyChanged();
            }
        }


        public double LeftPosition
        {
            get { return _leftPosition; }
            set
            {
                if (value < 0 || Math.Abs(_leftPosition - value) < 0.001) return;
                _leftPosition = value;
                OnPropertyChanged();
                this.Rectangle = CreateRectangle();
            }
        }

        public double TopPosition
        {
            get { return _topPosition; }
            set
            {
                if (value < 0 || Math.Abs(_topPosition - value) < 0.001) return;
                _topPosition = value;
                OnPropertyChanged();
                this.Rectangle = CreateRectangle();
            }
        }

        public double Width
        {
            get { return _width; }
            set
            {
                if (_width.Equals(value) || value < MinWidth) return;
                _width = value;
                OnPropertyChanged();
                this.Rectangle = CreateRectangle();
            }
        }

        public double Height
        {
            get { return _height; }
            set
            {
                if (_height.Equals(value)) return;
                _height = value;
                this.Rectangle = CreateRectangle();
                OnPropertyChanged();
            }
        }

        public double MinHeight
        {
            get { return _minHeight; }
        }

        public double MinWidth
        {
            get { return _minWidth; }
        }

        public bool IsSelected
        {
            get { return _isSelected; }
            set
            {
                _isSelected = value;
                OnPropertyChanged();
                if(_isSelected)
                    RaiseItemSelectedEvent();
            }
        }

        public int ZIndex
        {
            get { return _zIndex; }
            set
            {
                _zIndex = value;
                OnPropertyChanged();
            }
        }

        public Rectangle Rectangle 
        {
            get { return _rectangle; }
            set
            {
                if (_rectangle.Equals(value)) return;
                _rectangle = value;
                OnPropertyChanged();
                RaiseItemMovedEvent();
            }
        }

        public double ActualHeight
        {
            get { return _actualHeight; }
            set
            {
                if (value.Equals(_actualHeight)) return;
                _actualHeight = value;
                OnPropertyChanged();
                UpdateSizeToActualSize();
            }
        }

        public double ActualWidth
        {
            get { return _actualWidth; }
            set
            {
                if (_actualWidth.Equals(value)) return;

                _actualWidth = value;
                OnPropertyChanged();
                UpdateSizeToActualSize();
            }
        }


        public bool IsExpanded
        {
            get { return _isExpanded; }
            set
            {
                if (value.Equals(_isExpanded)) return;
                _isExpanded = value;
                OnPropertyChanged();
            }
        }

        public List<RelationshipViewModelBase> Relationships {get { return _relationships; }}

        public PointDistributor PointDistributor
        {
            get { return _pointDistributor ?? (_pointDistributor = new PointDistributor(this)); }
        }

        int IIndexedVertex.Index { get; set; }
        int IIndexedVertex.LowLink { get; set; }

        #endregion

        #region Private Methods

        private void UpdateSizeToActualSize()
        {           
            Height = ActualHeight;
            Width = ActualWidth;
        }

        private void RaiseItemSelectedEvent()
        {
            if(Selected != null)
                Selected(this,new EventArgs());
        }

        private void RaiseItemMovedEvent()
        {
			if (ItemMoved != null)
				ItemMoved(this, EventArgs.Empty);
		}

        protected virtual void EditItem()
        {
            
        }

        #endregion

        #region Public Methods

        /// <summary>
        /// Creates the rectangle with dimensions of this diagram item.
        /// </summary>
        /// <returns></returns>
        public Rectangle CreateRectangle()
        {
            return new Rectangle((int)LeftPosition,(int)TopPosition,(int)Width,(int)Height);
        }

        public Point[] GetAvailablePoints()
        {
            return Rectangle.GetPoints(Settings.Default.PathFindingSensitivity).ToArray();
        }

        public Point[] GetAvailablePoints(RelativePosition pointPosition)
        {
            if (Settings.Default.ExcludeUsedPoints)
            {
                var usedPoints = Relationships.Select(rel =>
                {
                    if (rel.Source.Equals(this))
                        return rel.SourcePoint;
                    return rel.TargetPoint;
                });
                return Rectangle.GetPoints(pointPosition, Settings.Default.PathFindingSensitivity * 3).Except(usedPoints).ToArray();
            }
            

            return Rectangle.GetPoints(pointPosition, Settings.Default.PathFindingSensitivity).ToArray();
        }

        public IEnumerable<T> GetOutgoingRelationships<T>() where T : RelationshipViewModelBase
        {
            return Relationships.OfType<T>().Where(rel => rel.Source.Equals(this));
        }
        public IEnumerable<RelationshipViewModelBase> GetOutgoingRelationships()
        {
            return Relationships.Where(rel => rel.Source.Equals(this));
        }

        public IEnumerable<T> GetIncomingRelationships<T>() where T : RelationshipViewModelBase
        {
            return Relationships.OfType<T>().Where(rel => rel.Target.Equals(this));
        }
        public IEnumerable<RelationshipViewModelBase> GetIncomingRelationships()
        {
            return Relationships.Where(rel => rel.Target.Equals(this));
        }

        #region IVertex, IIndexedVertex implementation

        public IEnumerable<IVertex> GetNeighbours(bool revert = false)
        {
            return GetIndexedNeighbours(revert);
        }

        public IEnumerable<IIndexedVertex> GetIndexedNeighbours(bool revert = false)
        {
            if (revert)
            {
                return Relationships.Where(rel => rel.Target.Equals(this)).Select(rel => rel.Source);
            }
            return Relationships.Where(rel => rel.Source.Equals(this)).Select(rel => rel.Target);
        }

        public IEnumerable<IVertex> GetNeighbours<T>(bool revert = false) where T : class
        {
            return GetIndexedNeighbours<T>(revert);
        }

        public IEnumerable<IIndexedVertex> GetIndexedNeighbours<T>(bool revert = false) where T :class
        {
            if (typeof (T).IsSubclassOf(typeof(RelationshipViewModelBase)))
            {
                if (revert)
                {
                    return Relationships.Where(rel => rel is T && rel.Target.Equals(this))
                        .Select(rel => rel.Source);
                }
                return Relationships.Where(rel => rel is T && rel.Source.Equals(this)).Select(rel => rel.Target);
            }
            //wrong generic type, so return non-generic version, which includes all relatioships
            return GetIndexedNeighbours(revert);
        }

        #endregion
        
        #endregion

        #region Commands

        public GenericCommand ItemSelectedCommand
        {
            get
            {
                return new GenericCommand(() => { IsSelected = true; });
            }
        }

        public ICommand EditItemCommand
        {
            get { return new GenericCommand(EditItem);}
        }


        public ICommand DeleteItemCommand { get
        {
            return new GenericCommand(() => Messenger.Send(new DeletedEntityDTO() {EntityToDelete = this}));
        } }

        #endregion

        public override bool Equals(object obj)
        {
            var item = obj as DiagramItemViewModelBase;
            if (item == null) return false;
            return item.GetHashCode() == GetHashCode();
        }

        public override int GetHashCode()
        {
            return _id.GetHashCode();
        }

        public event EventHandler Selected;
        public event EventHandler ItemMoved;
    }
}
