﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using UmlDiagramDesigner.Core;
using UmlDiagramDesigner.Core.Enums;
using UmlDiagramDesigner.DTO;
using UmlDiagramDesigner.Helpers;
using UmlDiagramDesigner.Models;
using UmlDiagramDesigner.Models.Attributes;
using UmlDiagramDesigner.ViewModels.Commands;
using UmlDiagramDesigner.ViewModels.DependentElements;
using UmlDiagramDesigner.ViewModels.DiagramRelationships;
using UmlDiagramDesigner.ViewModels.InputForms;
using UmlDiagramDesigner.Views.InputForms;

namespace UmlDiagramDesigner.ViewModels.DiagramItems
{
    [Associatable(SupportedAssocation = UmlAssociation.Dependency)]
    public class ViewDiagramItemViewModel : DiagramItemViewModelBase
    {
        private ViewEditViewModel _viewEditViewModel;
        private ObservableCollection<AttributeBySourceDTO> _attributes;

        public ViewDiagramItemViewModel()
        {
            Name = "ViewName";
            Attributes = new ObservableCollection<AttributeBySourceDTO>();
            Messenger.Register<DeletedEntityDTO>(this,dto => Attributes.Remove(att => att.Source.Equals(dto.EntityToDelete)));
        }

        #region Properties

        public ObservableCollection<AttributeBySourceDTO> Attributes
        {
            get { return _attributes; }
            set
            {
                _attributes = value;
                OnPropertyChanged();
            }
        }

        #endregion

        #region Commands

        public GenericCommand OpenEditWindowCommand
        {
            get { return new GenericCommand(OpenEditWindow); }
        }

        #endregion

        private void OpenEditWindow()
        {
            var groupedAttributes = GetGroupedAttributes();

            var window = new ViewEditWindow()
            {
                DataContext = _viewEditViewModel = new ViewEditViewModel(groupedAttributes)
                {
                    Name = this.Name
                }
            };

            _viewEditViewModel.InputCompleted += (sender, args) =>
            {
                Attributes.Clear();
                foreach (var attribute in _viewEditViewModel.GroupedAttributes.OrderByDescending(att => att.Attribute))
                {
                    if (attribute.IsSelected)
                    {
                        Attributes.Add(attribute);
                        attribute.Attribute.AttributeModified += AttributeOnModified;
                        attribute.Source.AttributeRemoved += OnAttributeRemoved;
                    }
                }

                Name = _viewEditViewModel.Name;
            };

            window.ShowDialog();
        }

        private IEnumerable<AttributeBySourceDTO> GetGroupedAttributes()
        {
            var sources = Relationships.OfType<DependencyViewModel>()
                .Where(relationship => relationship.Source.Equals(this) &&
                                       relationship.Target is ClassDiagramItemViewModel)
                .Select(dep => (ClassDiagramItemViewModel) dep.Target);

            List<AttributeBySourceDTO> groupedSource = Attributes != null
                ? new List<AttributeBySourceDTO>(Attributes)
                : new List<AttributeBySourceDTO>();

            foreach (var source in sources)
            {
                ClassDiagramItemViewModel currentSource = source;

                var grouped = (from a in source.Attributes
                    select new AttributeBySourceDTO()
                    {
                        Source = currentSource,
                        Attribute = a
                    });

                foreach (var attribute in grouped)
                {
                    if (!groupedSource.Contains(attribute))
                    {
                        groupedSource.Add(attribute);
                    }
                }
            }
            return groupedSource;
        }

        private void AttributeOnModified(object sender, EventArgs eventArgs)
        {
            var attribute = sender as AttributeViewModel;
            if (attribute == null) return;

            foreach (var attributeDto in Attributes)
            {
                if (attributeDto.Attribute == attribute)
                {
                    attributeDto.Attribute = attribute;
                }
            }
        }

        private void OnAttributeRemoved(object sender, AttributeViewModel attribute)
        {
            Attributes.Remove(attDto => attDto.Attribute == attribute);
        }

        public override string GetResourceStyleName()
        {
            return "ViewStyle";
        }
    }
}
