﻿using System.Collections.Generic;
using System.Windows;
using System.Windows.Input;
using UmlDiagramDesigner.Core;
using UmlDiagramDesigner.Core.Enums;
using UmlDiagramDesigner.Core.Graph;
using UmlDiagramDesigner.DTO;
using UmlDiagramDesigner.Helpers;
using UmlDiagramDesigner.Models;
using UmlDiagramDesigner.Models.PathFinders;
using UmlDiagramDesigner.Services;
using UmlDiagramDesigner.ViewModels.Commands;
using UmlDiagramDesigner.ViewModels.DependentElements;
using UmlDiagramDesigner.ViewModels.DiagramItems;

namespace UmlDiagramDesigner.ViewModels.DiagramRelationships
{
    public abstract class RelationshipViewModelBase : DiagramEntityViewModelBase, IAssociation<DiagramItemViewModelBase>
    {
        private Point _sourcePoint, _targetPoint;
        private IList<Point> _additionalPoints; 
        private DiagramItemViewModelBase _source, _target;
        protected IPathFinder PathFinder;
        protected IDialogService DialogService;
        private bool _isSelected;
        private MultiplicityType _sourceMultiplicity;
        private MultiplicityType _targetMultiplicity;
        private string _middleText;

        protected RelationshipViewModelBase()
        {
            PathFinder = new ShortestPathPathFinder();
            DialogService = new WpfDialogService();
            AdditionalPoints = new List<Point>();
            IsConnected = false;
            IsConnectable = false;
            CanBend = false;
        }

        #region Properties

        public DiagramItemViewModelBase Source 
        {
            get { return _source; }
            set
            {
                _source = value;
                if (_source != null)
                {
                    SourcePoint = new Point(_source.Rectangle.Location.X,_source.Rectangle.Location.Y);
                }
            }
        }

        public DiagramItemViewModelBase Target
        {
            get { return _target; }
            set
            {
                _target = value;
            }
        }

        public Point SourcePoint
        {
            get { return _sourcePoint; }
            set
            {
                if (value.X < 0 || value.Y < 0) return;
                _sourcePoint = value;
                OnPropertyChanged();
            }
        }

        public Point TargetPoint
        {
            get { return _targetPoint; }
            set
            {
                if (value.X < 0 || value.Y < 0) return;
                _targetPoint = value;
                OnPropertyChanged();
            }
        }

        public MultiplicityType SourceMultiplicity
        {
            get { return _sourceMultiplicity; }
            set { _sourceMultiplicity = value; OnPropertyChanged(); OnPropertyChanged("SourceText");}
        }

        public MultiplicityType TargetMultiplicity
        {
            get { return _targetMultiplicity; }
            set { _targetMultiplicity = value; OnPropertyChanged(); OnPropertyChanged("TargetText");}
        }

        public IList<Point> AdditionalPoints
        {
            get { return _additionalPoints;}
            set { _additionalPoints = value; OnPropertyChanged(); }
        }

        public string SourceText
        {
            get { return SourceMultiplicity.GetString(); }
        }

        public string TargetText
        {
            get { return TargetMultiplicity.GetString(); }
        }

        public string MiddleText
        {
            get { return _middleText; }
            set { _middleText = value; OnPropertyChanged();}
        }

        public double Opacity { get { return 1; } }

        public bool IsSelected
        {
            get { return _isSelected; }
            set
            {
                if (value.Equals(_isSelected)) return;
                _isSelected = value;
                OnPropertyChanged();
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether this instance is connected to another relationship.
        /// If it is, you don't have to draw end shape, if it isn't you have to.
        /// </summary>
        /// <value>
        /// <c>true</c> if this instance is connected; otherwise, <c>false</c>.
        /// </value>
        public bool IsConnected { get; set; }

        public bool IsConnectable { get; protected set; }

        public bool CanBend { get; protected set; }
         
        public AttributeViewModel SourceAttribute { get; set; }

        public AttributeViewModel TargetAttribute { get; set; }

        public bool IsHovered { get; set; }

        #endregion

        public ICommand EditRelationshipCommand
        {
            get { return new GenericCommand(OpenEditWindow); }
        }

        public ICommand RelationshipClickedCommand
        {
            get { return new GenericCommand(() => IsSelected = true); }   
        }

        public ICommand DeleteRelationshipCommand { get {return new GenericCommand(DeleteRelationship);}}

        #region Private Methods

        protected void DeleteRelationship()
        {
            if (Source != null)
                Source.Relationships.Remove(this);

            if (Target != null)
                Target.Relationships.Remove(this);

            var foreignKeyRelationship = this as IForeignKeyRelationship;
            if (foreignKeyRelationship != null)
            {
                var sourceEntity = Source as ClassDiagramItemViewModel;
                if (sourceEntity != null && sourceEntity.Attributes.Contains(foreignKeyRelationship.SourceForeignKey))
                {
                    sourceEntity.Attributes.Remove(foreignKeyRelationship.SourceForeignKey);
                }
            }

            Messenger.Send(new DeletedRelationshipDTO(this));
        }

        #endregion

        protected virtual void OpenEditWindow()
        {
            //nutny override v konkretnich tridach
        }

        public abstract override string ToString();
    }
}

