﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Threading.Tasks;
using System.Timers;
using System.Windows;
using System.Windows.Input;
using System.Windows.Threading;
using Microsoft.Win32;
using UmlDiagramDesigner.CodeGeneration;
using UmlDiagramDesigner.CodeGeneration.ConcreteGenerators;
using UmlDiagramDesigner.Core;
using UmlDiagramDesigner.DAL.Export;
using UmlDiagramDesigner.DAL.Serialization;
using UmlDiagramDesigner.Models;
using UmlDiagramDesigner.Services;
using UmlDiagramDesigner.ViewModels.Commands;
using UmlDiagramDesigner.ViewModels.DiagramItems;
using UmlDiagramDesigner.ViewModels.DiagramRelationships;
using UmlDiagramDesigner.Views;
using Timer = System.Threading.Timer;

namespace UmlDiagramDesigner.ViewModels
{
    public class MainPageViewModel : ViewModelBase
    {
        #region Variables

        private CanvasViewModel _canvasViewModel;
        private ToolboxViewModel _toolboxViewModel;

        private ISerializer _serializer;
        private IExporter _exporter;
        private IDialogService _dialogService;

        private readonly DiagramToDataModelConverter _diagramConverter;
        private SaveFileDialog _saveFileDialog;

        #endregion

        #region Constructors

        public MainPageViewModel()
        {
            _diagramConverter = new DiagramToDataModelConverter();
            Initialize();
        }

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets the canvas view model.
		/// </summary>
		/// <value>
		/// The canvas view model.
		/// </value>
		public CanvasViewModel CanvasViewModel
        {
            get { return _canvasViewModel; }
            set
            {
                _canvasViewModel = value;
                OnPropertyChanged();
                ChangeTracker.Instance.Track(value);
            }
        }

		/// <summary>
		/// Gets or sets the toolbox view model.
		/// </summary>
		/// <value>
		/// The toolbox view model.
		/// </value>
		public ToolboxViewModel ToolboxViewModel
        {
            get { return _toolboxViewModel; }
            set
            {
                _toolboxViewModel = value;
                OnPropertyChanged();
            }
        }

		/// <summary>
		/// Gets or sets a value indicating whether this object is currently importing data.
		/// It is used for disabling path finding when set to true.
		/// </summary>
		/// <value>
		/// <c>true</c> if this instance is importing data; otherwise, <c>false</c>.
		/// </value>
		public static bool IsImportingData { get; set; }

		#endregion

		#region Commands

		/// <summary>
		/// Creates new diagram by clearing all items from current one.
		/// </summary>
		/// <value>
		/// The create new diagram command.
		/// </value>
		public ICommand CreateNewDiagramCommand { get { return new GenericCommand(() =>
        {
            if (ChangeTracker.Instance.IsTracking && ChangeTracker.Instance.HasChanged(CanvasViewModel))
            {
                if (
                    _dialogService.Ask(
                        "There are some unsaved changes made to diagram, do you want to save it?"))
                {
                    ExportToJsonCommand.Execute(null);
                }
                else
                {
                    CanvasViewModel.Name = string.Empty;
                    CanvasViewModel.Items.Clear();
                }
            }
            else
            {
				CanvasViewModel.Name = string.Empty;
				CanvasViewModel.Items.Clear();
			}
        });}}

		/// <summary>
		/// Generates c# code from model.
		/// </summary>
		/// <value>
		/// The generate c sharp command.
		/// </value>
		public GenericCommand GenerateCSharpCommand
        {
            get
            {
                return new GenericCommand(async () =>
                {
                    await GenerateCodeAsync(new CSharpGenerator(new CSharpProperties()));
                });
            }
        }

		/// <summary>
		/// Generates java code from model.
		/// </summary>
		/// <value>
		/// The generate java command.
		/// </value>
		public GenericCommand GenerateJavaCommand
        {
            get
            {
                return new GenericCommand(async () =>
                {
                    await GenerateCodeAsync(new JavaGenerator(new JavaProperties()));
                });
            }
        }

		/// <summary>
		/// Generates smalltalk code from model.
		/// </summary>
		/// <value>
		/// The generate smalltalk command.
		/// </value>
		public GenericCommand GenerateSmalltalkCommand
        {
            get
            {
                return new GenericCommand(async () =>
                {
                    await GenerateCodeAsync(new SmallTalkGenerator(new SmallTalkProperties()));
                });
            }
        }

		/// <summary>
		/// Generates SQL code from model.
		/// </summary>
		/// <value>
		/// The generate SQL command.
		/// </value>
		public GenericCommand GenerateSqlCommand
        {
            get
            {
                return new GenericCommand(async () =>
                {
                    await GenerateCodeAsync(new SqlGenerator(new TablePerTypeGenerator(new MsSqlProperties())));
                });
            }
        }

		/// <summary>
		/// Exports visible part of diagram to chosen image format.
		/// </summary>
		/// <value>
		/// The export to image command.
		/// </value>
		public GenericCommand<FrameworkElement> ExportToImageCommand
        {
            get
            {
                return new GenericCommand<FrameworkElement>(canvas =>
                {
                    _exporter.Canvas = canvas;
                    _exporter.Export(null);
                });

            }
        }

		/// <summary>
		/// Exports diagram data to JSON format.
		/// </summary>
		/// <value>
		/// The export to json command.
		/// </value>
		public ICommand ExportToJsonCommand
        {
            get { return new GenericCommand(() =>
            {
                if (CanvasViewModel == null || CanvasViewModel.Items == null || !CanvasViewModel.Items.Any())
                {
                    _dialogService.Display(new Exception("There is no diagram to export."));
                    return;
                }

                _serializer.Serialize(CanvasViewModel.Items);
                ChangeTracker.Instance.Track(CanvasViewModel);
            });}
        }

		/// <summary>
		/// Imports data from JSON format.
		/// </summary>
		/// <value>
		/// The import from json command.
		/// </value>
		public ICommand ImportFromJsonCommand
        {
            get { return new GenericCommand(async () =>
            {

                if (ChangeTracker.Instance.IsTracking && ChangeTracker.Instance.HasChanged(CanvasViewModel))
                {
                    if (
                        _dialogService.Ask(
                            "There are some unsaved changes made to diagram, do you want to save it?"))
                    {
                        ExportToJsonCommand.Execute(null);
                    }
                    else
                    {
						await ImportDataFromJsonAsync();
                    }
                }
                else
                {
	                await ImportDataFromJsonAsync();
                }
            });}
        }

		/// <summary>
		/// Imports the data from json asynchronously. 
		/// Temporarily disables pathfinders, so they won't override placement of associations.
		/// </summary>
		/// <returns></returns>
		private async Task ImportDataFromJsonAsync()
	    {
		    IEnumerable<DiagramEntityViewModelBase> items = null;
		    DiagramEntityViewModelBase[] itemsArray = null;

		    await Task.Factory.StartNew(() => { items = _serializer.Deserialize(); });

		    if (items == null) return;

		    IsImportingData = true;

		    itemsArray = items as DiagramEntityViewModelBase[] ?? items.ToArray();
		    if (itemsArray.Any())
		    {
			    CanvasViewModel.Items =
				    new ObservableCollection<DiagramEntityViewModelBase>(itemsArray);
		    }

		    var timer = new System.Timers.Timer(500);
		    timer.Elapsed += (sender, args) =>
		    {
			    IsImportingData = false;
			    timer.Enabled = false;
		    };
		    timer.Enabled = true;
	    }

		/// <summary>
		/// Gets the open settings command.
		/// </summary>
		/// <value>
		/// The open settings command.
		/// </value>
		public ICommand OpenSettingsCommand {get {return new GenericCommand(() => new SettingsWindow().ShowDialog());}}


        public ICommand ExitApplicationCommand { get { return new GenericCommand(() =>
        {
            if (ChangeTracker.Instance.IsTracking && ChangeTracker.Instance.HasChanged(CanvasViewModel))
            {
                if (
                    _dialogService.Ask(
                        "There are some unsaved changes made to diagram, do you want to save it?"))
                {
                    ExportToJsonCommand.Execute(null);
                }
                else
                {
                    Application.Current.Shutdown();
                }
            }
            else
            {
				Application.Current.Shutdown();
			}
        });}}


		#endregion

		#region Private Methods

		/// <summary>
		/// Initializes this instance.
		/// </summary>
		private void Initialize()
        {
            CanvasViewModel = new CanvasViewModel();
            ToolboxViewModel = new ToolboxViewModel();

            _serializer = new JsonSerializer();
            _exporter = new ImageExporter();
            _dialogService = new WpfDialogService();
        }

		/// <summary>
		/// Creates the filter for file dialogs.
		/// </summary>
		/// <param name="generator">The generator.</param>
		/// <returns></returns>
		private string CreateFilter(GeneratorBase generator)
        {
            return string.Format("{0} |*{1}",generator.LanguageName,generator.FileExtension);
        }

		/// <summary>
		/// Generates the code asynchronously.
		/// </summary>
		/// <param name="generator">The generator.</param>
		/// <returns></returns>
		private async Task GenerateCodeAsync(GeneratorBase generator)
        {
            if (CanvasViewModel == null || CanvasViewModel.Items == null || !CanvasViewModel.Items.Any())
            {
                _dialogService.Display(new Exception("There is no diagram to generate code from."));
                return;
            }

            _saveFileDialog = new SaveFileDialog() { AddExtension = true, Filter = CreateFilter(generator) };

            var result = _saveFileDialog.ShowDialog();
            if (result.HasValue && result.Value)
            {
                try
                {
                    var model = _diagramConverter.Convert(CanvasViewModel);
                    model.Name = Path.GetFileNameWithoutExtension(_saveFileDialog.FileName);

                    var code = generator.GenerateCode(model);
                    await WriteStringToFile(_saveFileDialog.FileName, code);

                    OpenFile(_saveFileDialog.FileName);
                }
                catch (Exception e)
                {
                    _dialogService.Display(e);
                }
            }
        }

		/// <summary>
		/// Writes the string to file asynchronously. Replaces file if it exists.
		/// </summary>
		/// <param name="filePath">The file path.</param>
		/// <param name="text">The text.</param>
		/// <returns></returns>
		private async Task WriteStringToFile(string filePath, string text)
        {
            using (var writer = File.CreateText(filePath))
            {
                await writer.WriteAsync(text);
            }
        }

		/// <summary>
		/// Opens the file.
		/// </summary>
		/// <param name="filePath">The file path.</param>
		private void OpenFile(string filePath)
        {
            Process.Start(filePath);
        }

        #endregion
    }
}
