﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Security.Policy;
using System.Text;
using System.Threading.Tasks;
using UmlDiagramDesigner.CodeGeneration.Contracts;
using UmlDiagramDesigner.CodeGeneration.Enums;
using UmlDiagramDesigner.Core.Enums;
using UmlDiagramDesigner.Data;

namespace UmlDiagramDesigner.CodeGeneration.ConcreteGenerators
{
	/// <summary>
	/// Concrete generator for Java language.
	/// </summary>
	public class JavaGenerator : ObjectOrientedLanguageGenerator
    {
		/// <summary>
		/// Initializes a new instance of the <see cref="JavaGenerator"/> class.
		/// </summary>
		/// <param name="languageProperties">The language properties.</param>
		public JavaGenerator(IOopLanguageProperties languageProperties) : base(languageProperties)
        {
        }

        public override string LanguageName
        {
            get { return "Java"; }
        }

        public override string FileExtension
        {
            get { return ".java"; }
        }

        protected override IList<string> RequiredNamespaces
        {
            get
            {
                return new[]
                {
                    "java.lang.*",
                    "java.math.*",
                    "java.util.*",
                    "java.sql.*"
                };
            }
        }

        protected override string GenerateMainContent(Diagram diagram)
        {
            var sb = new StringBuilder();

            sb.AppendLine(GenerateNamespaceStart(diagram.Name));

            sb.AppendLine(GenerateNamespaceImports());

            foreach (var @class in diagram.Classes)
            {
                sb.AppendLine(GenerateClass(@class).InsertTabsAtStartOfEachLine());
            }

            sb.AppendLine(GenerateNamespaceEnd());

            return sb.ToString();
        }

        protected override string GenerateNamespaceStart(string namespaceName)
        {
            string package;
            if (LanguageProperties.Features.TryGetValue(LanguageFeature.Namespace, out package))
            {
                var endLine = LanguageProperties.Features[LanguageFeature.EndLineCharacter];
                return string.Format("{0} {1}{2}", package, namespaceName.ToLower(), endLine);
            }

            return string.Empty;
        }

        protected override string GenerateNamespaceEnd()
        {
            return string.Empty;
        }

        protected override string GenerateClass(Class @class)
        {
            var sb = new StringBuilder();

            sb.AppendLine(GenerateClassName(@class));
            sb.AppendLine(GenerateStartBlock());

            sb.AppendLine(GenerateFields(@class).InsertTabsAtStartOfEachLine());

            sb.AppendLine(GenerateConstructor(@class).InsertTabsAtStartOfEachLine());

            sb.AppendLine(GenerateGettersSetters(@class).InsertTabsAtStartOfEachLine());

            sb.AppendLine(GenerateEndBlock());

            return sb.ToString();
        }

        protected override string GenerateClassName(Class @class)
        {
            string result;

            if (@class.Ancestor == null)
            {
                result = string.Format("class {0}", @class.Name);
            }
            else
            {
                result = string.Format("class {0} {1} {2}", @class.Name,
                    LanguageProperties.Features[LanguageFeature.Inheritance], @class.Ancestor.Name);
            }

            return result;
        }

		/// <summary>
		/// Generates the fields.
		/// </summary>
		/// <param name="class">The class.</param>
		/// <returns></returns>
		private string GenerateFields(Class @class)
        {
            var sb = new StringBuilder();

            string privateAccessibility;
            LanguageProperties.Accessibilities.TryGetValue(Accessibility.Public, out privateAccessibility);

            if (@class.PrimaryKey != null)
            {
                sb.AppendLine(string.Format("{0} {1} {2}{3}", privateAccessibility ?? string.Empty,
                    LanguageProperties.DataTypes[@class.PrimaryKey.DataType],
                    @class.PrimaryKey.Name.ToCamelCase(),
                    GenerateLineEnd()));
            }

            if (@class.Attributes.Any())
            {
                foreach (var attribute in @class.Attributes)
                {
                    sb.AppendLine(string.Format("{0} {1} {2}{3}", privateAccessibility ?? string.Empty,
                        LanguageProperties.DataTypes[attribute.DataType],
                        attribute.Name.ToCamelCase(),
                        GenerateLineEnd()));
                }
            }

            if (@class.Aggregations.Any() || @class.Associations.Any() || @class.Compositions.Any())
            {
                foreach (var navigationAttribute in @class.Aggregations.Union(@class.Associations).Union(@class.Compositions))
                {
                    sb.AppendLine(string.Format("{0} {1} {2}{3}", privateAccessibility ?? string.Empty,
                        LanguageProperties.GenerateNavPropertyDataType(navigationAttribute),
                        DetermineNavPropertyName(navigationAttribute).ToCamelCase(),
                        GenerateLineEnd()));
                }
            }

            return sb.ToString();
        }

		/// <summary>
		/// Generates the getters setters.
		/// </summary>
		/// <param name="class">The class.</param>
		/// <returns></returns>
		private string GenerateGettersSetters(Class @class)
        {
            var sb = new StringBuilder();
            string comment, publicAccessibility;
            LanguageProperties.Features.TryGetValue(LanguageFeature.Comment, out comment);
            LanguageProperties.Accessibilities.TryGetValue(Accessibility.Public, out publicAccessibility);


            if (@class.PrimaryKey != null)
            {
                sb.AppendLine();
                //getter
                sb.AppendLine(string.Format("{0} {1} get{2}() {{ return this.{3}{4}}}", 
                    publicAccessibility ?? string.Empty,
                    LanguageProperties.DataTypes[@class.PrimaryKey.DataType],
                    @class.PrimaryKey.Name,
                    @class.PrimaryKey.Name.ToCamelCase(),
                    GenerateLineEnd()));
                //setter
                sb.AppendLine(string.Format("{0} void set{1}({2} {3}) {{ this.{4} = {5}{6}}}", 
                    publicAccessibility ?? string.Empty,
                    @class.PrimaryKey.Name,
                    LanguageProperties.DataTypes[@class.PrimaryKey.DataType],
                    @class.PrimaryKey.Name,
                    @class.PrimaryKey.Name.ToCamelCase(),
                    @class.PrimaryKey.Name,
                    GenerateLineEnd()));
            }

            if (@class.Attributes.Any())
            {
                sb.AppendLine();
                sb.AppendLine(string.Format("{0} Generated attributes.", comment));

                foreach (var attribute in @class.Attributes)
                {
                    //getter
                    sb.AppendLine(string.Format("{0} {1} get{2}() {{ return this.{3}{4}}}",
                        publicAccessibility ?? string.Empty,
                        LanguageProperties.DataTypes[attribute.DataType],
                        attribute.Name,
                        attribute.Name.ToCamelCase(),
                        GenerateLineEnd()));
                    //setter
                    sb.AppendLine(string.Format("{0} void set{1}({2} {3}) {{ this.{4} = {5}{6}}}",
                        publicAccessibility ?? string.Empty,
                        attribute.Name,
                        LanguageProperties.DataTypes[attribute.DataType],
                        attribute.Name,
                        attribute.Name.ToCamelCase(),
                        attribute.Name,
                        GenerateLineEnd()));
                    sb.AppendLine();
                }
            }

            if (@class.Aggregations.Any() || @class.Associations.Any() || @class.Compositions.Any())
            {
                sb.AppendLine();
                sb.AppendLine(string.Format("{0} Generated navigation properties.", comment));

                foreach (var navigationAttribute in @class.Aggregations.Union(@class.Associations).Union(@class.Compositions))
                {
                    //getter
                    sb.AppendLine(string.Format("{0} {1} get{2}() {{ return this.{3}{4}}}",
                        publicAccessibility ?? string.Empty,
                        LanguageProperties.GenerateNavPropertyDataType(navigationAttribute),
                        navigationAttribute.Name,
                        DetermineNavPropertyName(navigationAttribute).ToCamelCase(),
                        GenerateLineEnd()));
                    ////setter
                    sb.AppendLine(string.Format("{0} void set{1}({2} {3}) {{ this.{4} = {5}{6}}}",
                        publicAccessibility ?? string.Empty,
                        navigationAttribute.Name,
                        LanguageProperties.GenerateNavPropertyDataType(navigationAttribute),
                        navigationAttribute.Name,
                        DetermineNavPropertyName(navigationAttribute).ToCamelCase(),
                        navigationAttribute.Name,
                        GenerateLineEnd()));
                    sb.AppendLine();
                }
            }

            return sb.ToString();
        }

		/// <summary>
		/// Generates the constructor.
		/// </summary>
		/// <param name="class">The class.</param>
		/// <returns></returns>
		private string GenerateConstructor(Class @class)
        {
            string publicAccessibility;
            LanguageProperties.Accessibilities.TryGetValue(Accessibility.Public, out publicAccessibility);

            if (@class.Aggregations.Count == 0)
            {
                //just an empty constructor
                return string.Format("{0} {1}() {2}{3}", publicAccessibility ?? string.Empty, @class.Name,
                    GenerateStartBlock(), GenerateEndBlock());
            }

            var sb = new StringBuilder();
            sb.AppendLine(string.Format("{0} {1}(", publicAccessibility ?? string.Empty, @class.Name));
            foreach (var aggregation in @class.Aggregations)
            {
                sb.Append(string.Format("{0} {1},", LanguageProperties.GenerateNavPropertyDataType(aggregation),
                    DetermineNavPropertyName(aggregation).ToCamelCase()));
            }

            sb.Remove(sb.Length - 1, 1);//remove last ',' char

            sb.Append(string.Format(") \n {0}", GenerateStartBlock()));

            sb.Append(GenerateConstructorInitializations(@class.Aggregations).InsertTabsAtStartOfEachLine());

            sb.AppendLine(GenerateEndBlock());

            return sb.ToString();
        }

		/// <summary>
		/// Generates the constructor initializations.
		/// </summary>
		/// <param name="aggregations">The aggregations.</param>
		/// <returns></returns>
		private string GenerateConstructorInitializations(ICollection<ForeignKeyConstraint> aggregations)
        {
            var sb = new StringBuilder();

            foreach (var attribute in aggregations)
            {
                sb.AppendLine(string.Format("this.{0} = {1}{2}", 
                    DetermineNavPropertyName(attribute).ToCamelCase(),
                    DetermineNavPropertyName(attribute).ToCamelCase(),
                    GenerateLineEnd()));
            }

            return sb.ToString();
        }

    }
}
