﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using UmlDiagramDesigner.CodeGeneration.Contracts;
using UmlDiagramDesigner.CodeGeneration.Enums;
using UmlDiagramDesigner.Core.Enums;
using UmlDiagramDesigner.Data;

namespace UmlDiagramDesigner.CodeGeneration
{
    /// <summary>
    /// Base generator class. All concrete generators have to derive from it.
    /// Defines simple structure of generated document.
    /// </summary>
    public abstract class GeneratorBase
    {
		/// <summary>
		/// The <see cref="StringBuilder"/> which contains the final document.
		/// </summary>
		protected StringBuilder FinalStringBuilder;

        protected GeneratorBase(ILanguageProperties languageProperties)
        {
            FinalStringBuilder = new StringBuilder();
            LanguageProperties = languageProperties;
        }

        #region Properties

        public ILanguageProperties LanguageProperties { get; private set; }

		/// <summary>
		/// Gets the name of the language.
		/// </summary>
		/// <value>
		/// The name of the language.
		/// </value>
		public abstract string LanguageName { get; }

		/// <summary>
		/// Gets the file extension used for this language.
		/// </summary>
		/// <value>
		/// The file extension.
		/// </value>
		public abstract string FileExtension { get; }

		#endregion

		/// <summary>
		/// Generates the code from <param name="diagram"></param>.
		/// </summary>
		/// <param name="diagram">The diagram.</param>
		/// <returns></returns>
		public string GenerateCode(Diagram diagram)
        {
            FinalStringBuilder.Clear();

            FinalStringBuilder.AppendLine(GenerateOpening());
            
            FinalStringBuilder.AppendLine();

            FinalStringBuilder.AppendLine(GenerateMainContent(diagram));

            FinalStringBuilder.AppendLine(GenerateEnding());

            return FinalStringBuilder.ToString();
        }

		/// <summary>
		/// Generates the main content of the document.
		/// </summary>
		/// <param name="diagram">The diagram.</param>
		/// <returns></returns>
		protected abstract string GenerateMainContent(Diagram diagram);

		/// <summary>
		/// Generates the opening part of document. It contains only information data about generated code.
		/// </summary>
		/// <returns></returns>
		protected virtual string GenerateOpening()
        {
            var sb = new StringBuilder();

            string comment;
            if (LanguageProperties.Features.TryGetValue(LanguageFeature.Comment, out comment))
            {
                sb.AppendLine(string.Format("{0}***************************************************", comment));
                sb.AppendLine();
                sb.AppendLine(
                    string.Format("{0}This code was automatically generated to {1} language from UML diagram on {2}.", comment,
                        LanguageName, DateTime.Now.ToShortDateString()));
                sb.AppendLine(string.Format("{0}Use at your own risk.",comment));
                sb.AppendLine();
                sb.AppendLine(string.Format("{0}***************************************************", comment));
            }


            return sb.ToString();
        }

		/// <summary>
		/// Generates the ending of the document.
		/// </summary>
		/// <returns></returns>
		protected virtual string GenerateEnding()
        {
            return string.Empty;
        }
    }
}
