﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using UmlDiagramDesigner.Core.Graph;

namespace UmlDiagramDesigner.Core
{
	/// <summary>
	/// Implementation of <see cref="ICycleDetector"/> interface. It only detects cycles in graph.
	/// </summary>
	public class TarjansAlgorithm : ICycleDetector
    {
        private int _index = 0;
        private Stack<IIndexedVertex> _stack;

        private void StrongConnect(IIndexedVertex vertex, IList<IList<IIndexedVertex>> stronglyConnected)
        {
			if(vertex == null)
				throw new ArgumentNullException("vertex");
			if(stronglyConnected == null)
				throw new ArgumentNullException("stronglyConnected");

			vertex.Index = _index;
            vertex.LowLink = _index;
            _index++;
            _stack.Push(vertex);

            foreach (var neighbour in vertex.GetIndexedNeighbours())
            {
                if (neighbour.Index < 0)
                {
                    StrongConnect(neighbour, stronglyConnected);
                    vertex.LowLink = Math.Min(vertex.LowLink, neighbour.LowLink);
                }
                else if (_stack.Contains(neighbour))
                {
                    vertex.LowLink = Math.Min(vertex.LowLink, neighbour.Index);
                }
            }

            if (vertex.LowLink == vertex.Index)
            {
                var stronglyConnectedComponent = new List<IIndexedVertex>();
                IIndexedVertex tempVertex;
                do
                {
                    tempVertex = _stack.Pop();
                    stronglyConnectedComponent.Add(tempVertex);
                } while (tempVertex != null && !Equals(vertex, tempVertex));
                stronglyConnected.Add(stronglyConnectedComponent);
            }
        }

        public IReadOnlyCollection<IList<IIndexedVertex>> DetectCycles(IIndexedGraph graph)
        {
            IList<IList<IIndexedVertex>> stronglyConnected = new List<IList<IIndexedVertex>>();
            _stack = new Stack<IIndexedVertex>();
            _index = 0;
            ResetVertexIndexes(graph);

            foreach (var vertex in graph.GetVertices())
            {
                if (vertex.Index < 0)
                {
                    StrongConnect(vertex, stronglyConnected);
                }
            }

            return new ReadOnlyCollection<IList<IIndexedVertex>>(OmitIndependentComponents(stronglyConnected));
        }

        public IReadOnlyCollection<IList<IIndexedVertex>> DetectCycles<T>(
            IIndexedGraph graph) where T : class
        {
            IList<IList<IIndexedVertex>> stronglyConnected = new List<IList<IIndexedVertex>>();
            _stack = new Stack<IIndexedVertex>();
            _index = 0;
            ResetVertexIndexes(graph);

            foreach (var vertex in graph.GetVertices()) 
            {
                if (vertex.Index < 0)
                {
                    StrongConnect<T>(vertex, stronglyConnected);
                }
            }
            return new ReadOnlyCollection<IList<IIndexedVertex>>(
                OmitIndependentComponents(stronglyConnected));
        }

        private void StrongConnect<T>(
            IIndexedVertex vertex, IList<IList<IIndexedVertex>> stronglyConnected) where T : class
        {
            vertex.Index = _index;
            vertex.LowLink = _index;
            _index++;
            _stack.Push(vertex);

            foreach (var neighbour in vertex.GetIndexedNeighbours<T>())
            {
                if (neighbour.Index < 0)
                {
                    StrongConnect<T>(neighbour, stronglyConnected);
                    vertex.LowLink = Math.Min(vertex.LowLink, neighbour.LowLink);
                }
                else if (_stack.Contains(neighbour))
                {
                    vertex.LowLink = Math.Min(vertex.LowLink, neighbour.Index);
                }
            }

            if (vertex.LowLink == vertex.Index)
            {
                var stronglyConnectedComponent = new List<IIndexedVertex>();
                IIndexedVertex tempVertex;
                do
                {
                    tempVertex = _stack.Pop();
                    stronglyConnectedComponent.Add(tempVertex);
                } while (tempVertex != null && !Equals(vertex, tempVertex));
                stronglyConnected.Add(stronglyConnectedComponent);
            }
        }

		/// <summary>
		/// Omits the independent components (we only care about cycles in graph).
		/// </summary>
		/// <param name="components">The components.</param>
		/// <returns></returns>
		private IList<IList<IIndexedVertex>> OmitIndependentComponents(IEnumerable<IList<IIndexedVertex>> components)
        {
            var result = new List<IList<IIndexedVertex>>();
            //components with a single vertex are not cycles
            foreach (var stronglyConnected in components.Where(component => component.Count > 1))
            {
                result.AddRange((from vertex in stronglyConnected
                                 group vertex by vertex.LowLink //component can contain cycle and other single components
                                     into g
                                     where g.Count() > 1 //remove single components
                                     select g.Select(vertex => vertex) //pick only data (do not care for the key -> LowLink)
                                             .OrderBy(v => v.Index) //vertices are in reverse order, so order them by their Index
                                             .ToList()).ToList()); //put all those cycles to List and add them to final result
            }
            return result;
        }

		/// <summary>
		/// Resets the vertex indexes in <param name="graph"></param>.
		/// </summary>
		/// <param name="graph">The graph.</param>
		private static void ResetVertexIndexes(IIndexedGraph graph)
        {
            foreach (var vertex in graph.GetVertices())
            {
                vertex.Index = -1;
                vertex.LowLink = -1;
            }
        }
    }
}
